
def string_to_classification(string, sep=";"):
    """
    Converts a string of text into classifications.

    :param str string: The string that contains one or more classifications.
    :param str sep: The separator to split the `string` parameter on.
    :return: A list of AtlasClassification objects as dicts.
    :rtype: list(dict)
    """
    if string is None:
        return []
    # TODO: How do we bring in attributes if they're required?
    # Defaulting to NOT propagate the classification downstream.
    # Advanced users may decide to parse the classifications and update the
    # propagation themselves.
    results = [{"typeName": s.strip(), "attributes": {}, "propagate": False}
               for s in string.split(sep) if s.strip() != ""]
    return results


def columns_matching_pattern(row, starts_with, does_not_match=[]):
    """
    Takes in a json "row" and filters the keys to match the `starts_with`
    parameter.  In addition, it will remove any match that is included
    in the `does_not_match` parameter.

    :param dict row: A dictionary with string keys to be filtered.
    :param str starts_with:
        The required substring that all filtered results must start with.
    :param list(str) does_not_match:
        A list of key values that should be omitted from the results.
    :return: A dictionary that contains only the filtered results.
    :rtype: dict
    """
    candidates = {k: v for k, v in row.items() if str(
        k).startswith(starts_with)}
    for bad_key in does_not_match:
        if bad_key in candidates:
            candidates.pop(bad_key)
    candidates = {k[len(starts_with):].strip(): v for k,
                  v in candidates.items()}

    return candidates


def first_relationship_that_matches(end_def, end_def_type, end_def_name, relationship_typedefs):
    """
    Find the first relationship type that matches the end_def number, type
    and name from the provided typedefs.

    :param str end_def: Either 'endDef1' or 'endDef2'
    :param str end_def_type: The type within the end_def.
    :param str end_def_name:
        The name of the relationship attribute applied to the end def's entity.
        (e.g. columns, table, columnLineages, query)
    :param list(dict) relationship_typedefs:
        A list of dictionaries that follow the relationship type defs.
    :raises ValueError:
        An relationship dict was not found in the provided typedefs.
    :return: The matching relationship type definition.
    :rtype: dict
    """
    output = None
    for typedef in relationship_typedefs:
        if ((end_def in typedef) and
            (typedef[end_def]["type"] == end_def_type) and
                (typedef[end_def]["name"] == end_def_name)):
            output = typedef

    if output is None:
        raise ValueError(
            "Unable to find a relationship type that matches: {endDef} "
            "with type {end_def_type} and the name {end_def_name} from "
            "the {num_defs} provided."
            .format(
                endDef=end_def, end_def_type=end_def_type,
                end_def_name=end_def_name, num_defs=len(relationship_typedefs)
            )
        )

    return output


def first_entity_matching_attribute(attribute, value, atlas_entities):
    """
    Return the first entity in a list that matches the passed in attribute
    and value.

    :param str attribute: The name of the attribute to search on each
        atlas entity.
    :param str value: The value of the attribute to search on each
        atlas entity.
    :param atlas_entities: The list of atlas entities to search over.
    :type atlas_entities: list(:class:`~pyapacheatlas.core.entity.AtlasEntity`)
    :raises ValueError:
        A matching entity was not found in the provided entities.
    :return: The atlas entity that maches the attribute and value.
    :rtype: :class:`~pyapacheatlas.core.entity.AtlasEntity`
    """
    output = None
    for entity in atlas_entities:
        if attribute in entity.attributes:
            if entity.attributes[attribute] == value:
                output = entity
                break
    if output is None:
        raise ValueError(
            "Unable to find an entity that matches the value of {value} in "
            "an attribute of {attribute} from the {num_entities} provided."
            .format(
                value=value, attribute=attribute,
                num_entities=len(atlas_entities)
            )
        )
    return output


def first_process_containing_io(input_name, output_name, atlas_entities):
    """
    Return the first entity in a list that contains the inputs and outputs.
    If input_name and output_name are the only or one of the many inputs and
    outputs (respectively), it will count as a match.

    :param str inputs:
        The qualified name of an atlas entity or a '*' wildcard
        that matches to any input (empty list or filled in values)
    :param str outputs: The qualified name of an atlas entity.
    :param atlas_entities: The list of atlas entities to search over.
    :type atlas_entities: list(:class:`~pyapacheatlas.core.entity.AtlasEntity`)
    :raises ValueError:
        A matching entity was not found in the provided entities.
    :return: The atlas entity that maches.
    :rtype: :class:`~pyapacheatlas.core.entity.AtlasEntity`
    """
    output_entity = None
    for entity in atlas_entities:
        input_matches = False
        output_matches = False
        if "inputs" in entity.attributes and "outputs" in entity.attributes:
            num_inputs = len(entity.attributes["inputs"])
            num_outputs = len(entity.attributes["outputs"])
            input_matches = ((input_name == "*") or  # Wildcard
                             ((input_name is None) and (num_inputs == 0)) or
                             ((input_name is not None) and (num_inputs > 0) and
                              (any([e["qualifiedName"] ==
                                    input_name for e in entity.inputs ]))
                              )
                             )
            output_matches = (
                ((output_name is None) and (num_outputs == 0)) or
                ((output_name is not None) and (num_outputs > 0) and
                    (any([e["qualifiedName"] == output_name for e in entity.outputs ]))
                 )
            )
        if input_matches and output_matches:
            output_entity = entity

    if output_entity is None:
        num_entities = len(atlas_entities)
        raise ValueError(
            "Unable to find a process that includes input "
            f"qualified names: {input_name} and output "
            f"qualified names: {output_name} from the "
            f"{num_entities} provided."
            .format(
                input_name=input_name, output_name=output_name,
                num_entities=num_entities
            )
        )

    return output_entity


def from_process_lookup_col_lineage(process_name, atlas_entities, relationship_typedefs):
    """
    Given a process name, find the first entity that matches that process name.
    Given that process entity, find the first relationship that contains that
    process type as an endDef2.

    :param str process_name: The name of the process that you want to find.
    :param atlas_entities: The list of atlas entities to search over.
    :type atlas_entities: list(:class:`~pyapacheatlas.core.entity.AtlasEntity`)
    :param list(dict) relationship_typedefs:
        The list of relationship type definitions to extract from.
    :raises ValueError:
        A matching entity or matching relationship type was not found in
        the provided entities or type def.
    :return: The endDef1 type from the discovered process relationship type.
    :rtype: str
    """
    target_entity = first_entity_matching_attribute(
        "name", process_name, atlas_entities)
    # TODO: Make "columnLineages" dynamic so that you can control which
    # attribute you're searching for

    lineage_relationship = first_relationship_that_matches(
        end_def="endDef2",
        end_def_type=target_entity.typeName,
        end_def_name="columnLineages",
        relationship_typedefs=relationship_typedefs
    )

    column_lineage_type = lineage_relationship["endDef1"]["type"]

    return column_lineage_type


def _make_col_qual_name(col_name, parent_table_name):
    """
    Generate a standardized qualified name for column entities.

    :param str col_name: The column name.
    :param str parent_table_name: The parent table name.

    :return: The qualified name for the column.
    :rtype: str
    """
    return f"{parent_table_name}#{col_name}"



from ..core import typedef


def column_lineage_scaffold(datasource,
                            use_column_mapping=False,
                            column_attributes=[],
                            table_attributes=[],
                            table_column_relationship_attributes=[],
                            column_lineage_attributes=[],
                            table_process_attributes=[],
                            column_lineage_process_attributes=[],
                            table_process_column_lineage_relationship_attributes=[]
                            ):
    """
    Create a base set of type definitions that adhere to the Hive Bridge
    style of Atlas Column Lineage.

    :param str datasource: The name of the data source. Acts as a prefix
        for all other type defs.
    :param bool use_column_mapping: If True, add the columnMapping attribute
        to the table process.
    :param list(dict), optional column_attributes:
        Attribute Defs to add to the column entity type.
    :param list(dict), optional table_attributes:
        Attribute Defs to add to the table entity type.
    :param list(dict), optional table_column_relationship_attributes:
        Attribute Defs to add to the table column relationship type.
    :param list(dict), optional column_lineage_attributes:
        Attribute Defs to add to the column lineage process entity type.
    :param list(dict), optional table_process_attributes:
        Attribute Defs to add to the table process entity type.
    :param list(dict), optional column_lineage_process_attributes:
        Attribute Defs to add to the column lineage process entity type.
    :param list(dict), optional table_process_column_lineage_relationship_attributes:
        Attribute Defs to add to the table process column lineage
        relationship type.
    """

    src_table_columns_typeName = "{}_table_columns".format(datasource)

    # TODO: Create all combinations of datasource
    # Define {datasource}_column
    column_entity = typedef.EntityTypeDef(
        name="{}_column".format(datasource),
        superTypes=["DataSet"],
        attributeDefs=column_attributes
    )
    # Define {datasource}_table
    table_entity = typedef.EntityTypeDef(
        name="{}_table".format(datasource),
        superTypes=["DataSet"],
        attributeDefs=table_attributes,
        relationshipAttributeDefs=[],
        options={"schemaElementsAttribute": "columns"}
    )
    # Define {datasource}_table_columns relationship ()
    table_column_relationship = typedef.RelationshipTypeDef(
        name=src_table_columns_typeName,
        attributeDefs=table_column_relationship_attributes,
        relationshipCategory="COMPOSITION",
        endDef1=typedef.ParentEndDef(name="columns",typeName=table_entity.name),
        endDef2=typedef.ChildEndDef(name="table",typeName=column_entity.name)
    )

    # Define {datasource}_column_lineage
    column_lineage_process_entity = typedef.EntityTypeDef(
        name="{}_column_lineage".format(datasource),
        superTypes=["Process"],
        attributeDefs=((column_lineage_process_attributes or []) +
                       [
            typedef.AtlasAttributeDef(
                name="dependencyType",
                isOptional=True,
                valuesMinCount=1,
                valuesMaxCount=1
            ).to_json(),
            typedef.AtlasAttributeDef(
                name="expression"
            ).to_json()
        ]
        )
    )

    # Define {datasource}_process
    table_process_entity = typedef.EntityTypeDef(
        name="{}_process".format(datasource),
        superTypes=["Process"],
        attributeDefs=table_process_attributes
    )
    if use_column_mapping:
        table_process_entity.attributeDefs.append(
            typedef.AtlasAttributeDef(
                name="columnMapping"
            ).to_json()
        )

    # Define {datasource}_process_column_lineage
    table_process_column_lineage_relationship = typedef.RelationshipTypeDef(
        name="{}_process_column_lineage".format(datasource),
        relationshipCategory="COMPOSITION",
        attributeDefs=table_process_column_lineage_relationship_attributes,
        endDef1=typedef.ChildEndDef(name="query",
            typeName=column_lineage_process_entity.name,
            isLegacyAttribute=True),
        endDef2=typedef.ParentEndDef(name="columnLineages",
            typeName=table_process_entity.name,
            isLegacyAttribute=False)
    )

    # Output composite entity
    output = {
        "entityDefs": [
            column_entity.to_json(),
            table_entity.to_json(),
            column_lineage_process_entity.to_json(),
            table_process_entity.to_json()
        ],
        "relationshipDefs": [
            table_column_relationship.to_json(),
            table_process_column_lineage_relationship.to_json()
        ]
    }
    return output



from string import ascii_uppercase

from openpyxl import Workbook
from openpyxl import load_workbook

from .reader import Reader, ReaderConfiguration


class ExcelConfiguration(ReaderConfiguration):
    """
    A configuration utility to understand how your Excel file is structured.

    You must have a "Columns" and "Tables" sheet.  The name is configurable
    with the column_sheet and table_sheet properties.

    The Columns sheet must contain a "Source/Target" Column and Table header.
    Optionally, a Classifications column can be provided for each
    Source/Target.

    The Tables sheet must contain a "Source/Target" Table and Type along with
    a Process Name and Process Type.  The Process is related to the mechanism
    by which source becomes the target (e.g. a Stored Procedure or Query).

    :param str column_sheet: Defaults to "Columns"
    :param str table_sheet: Defaults to "Tables"
    :param str entityDef_sheet: Defaults to "EntityDefs"
    :param str entity_source_prefix:
        Defaults to "source" and represents the prefix of the columns
        in Excel to be considered related to the source table or column.
    :param str entity_target_prefix:
        Defaults to "target" and represents the prefix of the columns
        in Excel to be considered related to the target table or column.
    :param str entity_process_prefix:
        Defaults to "process" and represents the prefix of the columns
        in Excel to be considered related to the table process.
    :param str column_transformation_name:
        Defaults to "transformation" and identifies the column that
        represents the transformation for a specific column.
    """

    def __init__(self, column_sheet="ColumnsLineage",
                 table_sheet="TablesLineage",
                 entityDef_sheet="EntityDefs", bulkEntity_sheet="BulkEntities",
                 updateLineage_sheet="UpdateLineage",
                 **kwargs):
        super().__init__(**kwargs)
        # Required attributes:
        # qualifiedName, column, transformation, table
        self.column_sheet = column_sheet
        self.table_sheet = table_sheet
        self.entityDef_sheet = entityDef_sheet
        self.bulkEntity_sheet = bulkEntity_sheet
        self.updateLineage_sheet = updateLineage_sheet


class ExcelReader(Reader):
    """
    Read in Excel files that follow the excel template tab structure.
    Expects an :class:`~pyapacheatlas.readers.excel.ExcelConfiguration` object
    to determine the naming conventions of tabs and headers.
    """
    @staticmethod
    def _parse_spreadsheet(worksheet):
        """
        Standardizes the excel worksheet into a json format.

        :param openpyxl.workbook.Workbook worksheet:
            A worksheet class from openpyxl.
        :return:
            The standardized version of the excel spreadsheet in json form.
        :rtype: list(dict(str,str))
        """
        # Standardize the column header
        # TODO: Allow for skip lines
        column_headers = list(
            zip(
                range(0, len(worksheet[1])),
                [str(c.value).strip() for c in worksheet[1]]
            )
        )

        output = []
        for row in worksheet.iter_rows(
                min_row=2, max_row=worksheet.max_row,
                min_col=1, max_col=worksheet.max_column):
            output.append(
                {k: row[idx].value for idx, k in column_headers})

        return output

    def parse_bulk_entities(self, filepath):
        """
        Generate a set of entities from an excel template file.

        :param str filepath:
            The xlsx file that contains your table and columns.
        :return: An AtlasTypeDef with entityDefs for the provided rows.
        :rtype: dict(str, list(dict))
        """
        wb = load_workbook(filepath)
        # A user may omit the entityDef_sheet by providing the
        # config with None
        sheetIsNotPresent = self.config.bulkEntity_sheet not in wb.sheetnames
        if self.config.bulkEntity_sheet and sheetIsNotPresent:
            raise KeyError("The sheet {} was not found".format(
                self.config.entityDef_sheet))

        output = dict()

        if self.config.bulkEntity_sheet:
            bulkEntity_sheet = wb[self.config.bulkEntity_sheet]
            json_bulkEntities = ExcelReader._parse_spreadsheet(
                bulkEntity_sheet)
            bulkEntities_generated = super().parse_bulk_entities(
                json_bulkEntities)
            output.update(bulkEntities_generated)

        wb.close()

        return output

    def parse_entity_defs(self, filepath):
        """
        Read a given excel file that conforms to the excel atlas template and
        parse the type def tab(s) into a set of entity defs that can be
        uploaded.

        Currently, only entityDefs are supported.

        :param str filepath:
            The xlsx file that contains your table and columns.
        :return: An AtlasTypeDef with entityDefs for the provided rows.
        :rtype: dict(str, list(dict))
        """
        wb = load_workbook(filepath)
        # A user may omit the entityDef_sheet by providing the config with None
        sheetIsNotPresent = self.config.entityDef_sheet not in wb.sheetnames
        if self.config.entityDef_sheet and sheetIsNotPresent:
            raise KeyError("The sheet {} was not found".format(
                self.config.entityDef_sheet))

        output = dict()

        # Getting entityDefinitions if the user provided a name of the sheet
        if self.config.entityDef_sheet:
            entityDef_sheet = wb[self.config.entityDef_sheet]
            json_entitydefs = ExcelReader._parse_spreadsheet(entityDef_sheet)
            entityDefs_generated = super().parse_entity_defs(json_entitydefs)
            output.update(entityDefs_generated)

        wb.close()
        # TODO: Add in classificationDefs and relationshipDefs
        return output

    def parse_column_lineage(self, filepath, atlas_entities, atlas_typedefs, use_column_mapping=False):
        """
        Read a given excel file that conforms to the excel atlas template and
        parse the columns into column lineages.
        Requires that the relationship attributes are already defined in the
        provided atlas type defs.

        Infers column type from the target table type and an assumed "columns"
        relationship attribute on the table type.

        Infers the column lineage process based on the provided table process
        (provided in the template's table excel sheet).  Looks for the first
        relationship type def with an endDef2 of `columnLineages`.

        :param str filepath:
            The xlsx file that contains your table and columns.
        :param list() atlas_entities:
            A list of AtlasEntity objects representing
        :param dict(str,list(dict)) atlas_typedefs:
            The results of requesting all type defs from Apache Atlas,
            including entityDefs, relationshipDefs, etc.  relationshipDefs
            are the only values used.
        :param bool use_column_mapping:
            Should the table processes include the columnMappings attribute
            that represents Column Lineage in Azure Data Catalog.
            Defaults to False.
        :return:
            A list of Atlas Entities representing the spreadsheet's
            inputs as their json dicts.
        :rtype: list(:class:`~pyapacheatlas.core.entity.AtlasEntity`)
        """

        wb = load_workbook(filepath)

        if self.config.column_sheet not in wb.sheetnames:
            raise KeyError("The sheet {} was not found".format(
                self.config.column_sheet))

        # Getting column entities
        column_sheet = wb[self.config.column_sheet]
        json_columns = ExcelReader._parse_spreadsheet(column_sheet)

        entities = super().parse_column_lineage(
            json_columns,
            atlas_entities,
            atlas_typedefs,
            use_column_mapping=use_column_mapping
        )

        wb.close()

        return entities

    def parse_table_lineage(self, filepath):
        """
        Read a given excel file that conforms to the excel atlas template and
        parse the tables and processes table lineages.
        Requires that the relationship attributes are already defined in the
        provided atlas type defs.

        :param str filepath:
            The xlsx file that contains your table and columns.
        :return:
            A list of Atlas Entities representing the spreadsheet's inputs
            as their json dicts.
        :rtype: list(:class:`~pyapacheatlas.core.entity.AtlasEntity`)
        """

        wb = load_workbook(filepath)

        entities = []

        if self.config.table_sheet not in wb.sheetnames:
            raise KeyError("The sheet {} was not found".format(
                self.config.table_sheet))

        # Getting table entities
        table_sheet = wb[self.config.table_sheet]
        json_sheet = ExcelReader._parse_spreadsheet(table_sheet)
        entities = super().parse_table_lineage(json_sheet)

        wb.close()

        return entities

    def parse_lineages(self, filepath, atlas_typedefs, use_column_mapping=False):
        """
        Read a given excel file that conforms to the excel atlas template and
        parse the tables, processes, and columns into table and column
        lineages. Requires that the relationship attributes are already
        defined in the provided atlas type defs.

        Infers column type from the target table type and an assumed "columns"
        relationship attribute on the table type.

        Infers the column lineage process based on the provided table process
        (provided in the template's table excel sheet).  Looks for the first
        relationship type def with an endDef2 of `columnLineages`.

        :param str filepath:
            The xlsx file that contains your table and columns.
        :param dict(str,list(dict)) atlas_typedefs:
            The results of requesting all type defs from Apache Atlas,
            including entityDefs, relationshipDefs, etc.  relationshipDefs are
            the only values used.
        :param bool use_column_mapping:
            Should the table processes include the columnMappings attribute
            that represents Column Lineage in Azure Data Catalog.
            Defaults to False.
        :return:
            A list of Atlas Entities representing the spreadsheet's inputs
            as their json dicts.
        :rtype: list(dict)
        """
        entities = []

        table_entities = self.parse_table_lineage(filepath)
        entities.extend(table_entities)

        # Modifies table_entities if use_column_mapping is True
        column_entities = self.parse_column_lineage(
            filepath,
            table_entities,
            atlas_typedefs,
            use_column_mapping
        )
        entities.extend(column_entities)

        output = [e.to_json() for e in entities]

        return output

    def parse_update_lineage(self, filepath):
        """
        Read a given excel file that conforms to the excel atlas template and
        parse the (default) UpdateLineage table into existing process entities. 

        Assumes these process entities and any referenced entity exists.

        Leave the qualifiedName cell blank on source or target to leave the
        existing input or output (respectively) unchanged.

        Use 'N/A' in the qualifiedName on source or target to 'destroy' the
        existing input or output and overwrite with an empty list.

        :param str filepath:
            The xlsx file that contains your table and columns.
        :return:
            A list of Atlas Process entities representing the spreadsheet's
            contents.
        :rtype: list(dict)
        """
        wb = load_workbook(filepath)

        entities = []

        if self.config.updateLineage_sheet not in wb.sheetnames:
            raise KeyError("The sheet {} was not found".format(
                self.config.updateLineage_sheet))

        # Getting table entities
        updateLineage_sheet = wb[self.config.updateLineage_sheet]
        json_sheet = ExcelReader._parse_spreadsheet(updateLineage_sheet)
        entities = super().parse_update_lineage(json_sheet)

        wb.close()

        return entities

    @staticmethod
    def _update_sheet_headers(headers, worksheet):
        """
        For the given worksheet, make the first row equal to the list passed
        in as the headers.

        :param list headers: A list of column headers to use for this sheet.
        :param worksheet:
            The worksheet you are updating.
        :type worksheet:
            :class:`~openpyxl.worksheet.worksheet.Worksheet`
        """
        for idx, val in enumerate(headers):
            # TODO: Not the best way once we get past
            # 26 columns in the template
            active_column = ascii_uppercase[idx]
            active_value = headers[idx]
            active_cell = "{}1".format(active_column)
            worksheet[active_cell] = active_value
            worksheet.column_dimensions[active_column].width = len(
                active_value)

    @staticmethod
    def make_template(filepath):
        """
        Generate an Excel template file and write it out to the given filepath.

        :param str filepath: The file path to store an XLSX file with the
            template Tables and Columns sheets.
        """
        wb = Workbook()
        columnsSheet = wb.active
        columnsSheet.title = "ColumnsLineage"
        tablesSheet = wb.create_sheet("TablesLineage")
        entityDefsSheet = wb.create_sheet("EntityDefs")
        bulkEntitiesSheet = wb.create_sheet("BulkEntities")
        updateLineageSheet = wb.create_sheet("UpdateLineage")

        ExcelReader._update_sheet_headers(
            Reader.TEMPLATE_HEADERS["ColumnsLineage"], columnsSheet
        )
        ExcelReader._update_sheet_headers(
            Reader.TEMPLATE_HEADERS["TablesLineage"], tablesSheet
        )
        ExcelReader._update_sheet_headers(
            Reader.TEMPLATE_HEADERS["EntityDefs"], entityDefsSheet
        )
        ExcelReader._update_sheet_headers(
            Reader.TEMPLATE_HEADERS["BulkEntities"], bulkEntitiesSheet
        )
        ExcelReader._update_sheet_headers(
            Reader.TEMPLATE_HEADERS["UpdateLineage"], updateLineageSheet
        )

        wb.save(filepath)
        wb.close()



import base64

from .base import AtlasAuthBase


class BasicAuthentication(AtlasAuthBase):
    """
    Authenticates using basic authentication.  Primarily used for testing
    on local
    """

    def __init__(self, username, password):
        """
        :param str username: The basic auth username.
        :param str username: The basic auth password.
        """
        super().__init__()
        self._username = username
        self._password = password

    def get_authentication_headers(self):
        """
        Provide the authentication headers for HTTP requests.
        :return: The authorization headers.
        :rtype: dict(str, str)
        """
        byte_user_pass = bytes("{u}:{pw}".format(
            u=self._username, pw=self._password), "utf-8")
        b64_str_userpass = base64.standard_b64encode(
            byte_user_pass).decode("utf-8")
        return {"Authorization": "Basic {}".format(b64_str_userpass)}



from abc import ABC
from abc import abstractmethod


class AtlasAuthBase(ABC):
    """
    The base class for authentication to your Apache Atlas server.
    """

    def __init__(self):
        super().__init__()

    @abstractmethod
    def get_authentication_headers(self, **kwargs):
        """
        Provide the authentication headers for HTTP requests.
        :return: The authorization headers.
        :rtype: dict(str, str)
        """
        raise NotImplementedError



from collections import namedtuple
import json
import warnings

EntityField = namedtuple("EntityField", ["name", "isOptional"])


class WhatIfValidator():
    """
    Provides a simple way to validate that your entities will successfully
    upload.  Provides functions to validate the type, check if required
    attributes are missing, and check if superfluous attributes are included.

    :param dict type_defs:
        The list of type definitions to be validated against.  Should be
        in the form of an AtlasTypeDef composite wrapper.
    :param list(dict) existing_entities:
        The existing entities that should be validated against.
    """
    ASSET_ATTRIBUTES = ["name", "description", "owner"]
    REFERENCABLE_ATTRIBUTES = ["qualifiedName"]

    ATLAS_MODEL = {
        "ASSET": ["name", "description", "owner"],
        "REFERENCABLE": ["qualifiedName"],
        "PROCESS": (["inputs", "outputs"] +
                    ASSET_ATTRIBUTES + REFERENCABLE_ATTRIBUTES),
        "DATASET": ASSET_ATTRIBUTES + REFERENCABLE_ATTRIBUTES,
        "INFRASTRUCTURE": ASSET_ATTRIBUTES + REFERENCABLE_ATTRIBUTES
    }

    def __init__(self, type_defs={}, existing_entities=[]):
        if len(type_defs) == 0 and len(existing_entities) == 0:
            warnings.warn(
                "WARNING: Provided type_defs and existing_entities are empty.  All validations will pass.")

        self.classification_defs = type_defs.get("classificationDefs", [])
        self.entity_defs = type_defs.get("entityDefs", [])
        # Create a dict of all entities by name, then find the name of all
        # attributes and whether they are optional
        entity_fields = {}
        for e in self.entity_defs:
            entity_fields[e["name"]] = []
            for attr in e.get("attributeDefs", []):
                ef = EntityField(attr.get("name"), attr.get("isOptional"))
                entity_fields[e["name"]].append(ef)

        # Adding Qualified Name to the set of valid fields as it doesn't
        # show up in the entity type def
        self.entity_valid_fields = {k: set(
            [field.name for field in v] + ["qualifiedName"])
            for k, v in entity_fields.items()}
        # Adding Qualified Name to the set of required entity
        # fields as it doesn't show up in the entity type def
        self.entity_required_fields = {
            k: set([field.name for field in v if not field.isOptional] + [
                "qualifiedName"]) for k, v in entity_fields.items()
        }

        self.existing_entities = set(
            [e.get("attributes", {}).get("qualifiedName") for
             e in existing_entities])

        self.enum_defs = type_defs.get("enumDefs", [])
        self.relationship_defs = type_defs.get("relationshipDefs", [])
        self.struct_defs = type_defs.get("structDefs", [])

    def entity_type_exists(self, entity):
        """
        Validate that the entity's type is an actual entity definition.

        :param dict entity:
        :return: Whether the entity matches the list of known entity types.
        :rtype: bool
        """
        current_type = entity["typeName"]
        if current_type in self.entity_valid_fields:
            return True
        else:
            return False

    def entity_missing_attributes(self, entity):
        """
        Check if the entity is missing required attributes.

        :param dict entity:
        :param dict type_def:
        :return: Whether the entity matches the list of known entity types.
        :rtype: bool
        """

        current_attributes = set(entity.get("attributes", {}).keys())
        required_attributes = set(
            self.entity_required_fields[entity["typeName"]])
        missing_attributes = required_attributes.difference(current_attributes)
        if len(missing_attributes) > 0:
            return missing_attributes
        else:
            return False

    def entity_has_invalid_attributes(self, entity):
        """
        Check if the entity is using attributes that are not defined
        on the type.

        :param dict entity:
        :return: Whether the entity matches the list of known entity types.
        :rtype: bool
        """
        current_attributes = set(entity.get("attributes", {}).keys())
        valid_attributes = set(self.entity_valid_fields[entity["typeName"]])
        # Append inherited attributes:
        _entity_type = entity["typeName"]
        # Assuming only one entity matches and only one super type
        super_type = [e["superTypes"]
                      for e in self.entity_defs
                      if e["name"] == _entity_type][0][0].upper()
        if super_type in self.ATLAS_MODEL:
            valid_attributes = valid_attributes.union(
                self.ATLAS_MODEL[super_type])
        invalid_attributes = current_attributes.difference(valid_attributes)

        if len(invalid_attributes) > 0:
            return invalid_attributes
        else:
            return False

    def entity_would_overwrite(self, entity):
        """
        Based on the qualified name attributes, does the provided
        entity exist in the entities provided to the What If Validator?

        :param dict entity:
        :return: Whether the entity matches an existing entity.
        :rtype: bool
        """
        current_qualifiedName = entity.get(
            "attributes", {}).get("qualifiedName")
        if current_qualifiedName is None:
            raise KeyError("Entity has no qualifiedName:\n{}".format(
                json.dumps(entity, indent=1)))

        if current_qualifiedName in self.existing_entities:
            return True
        else:
            return False

    def validate_entities(self, entities):
        """
        Provide a report of invalid entities.  Includes TypeDoesNotExist,
        UsingInvalidAttributes, and MissingRequiredAttributes.

        :param list(dict) entities: A list of entities to validate.
        :return: A dictionary containing counts values for the above values.
        :rtype: dict
        """
        report = {"TypeDoesNotExist": [], "UsingInvalidAttributes": {},
                  "MissingRequiredAttributes": {}}

        for entity in entities:
            if not self.entity_type_exists(entity):
                report["TypeDoesNotExist"].append(entity["guid"])
                # If it's an invalid type, we have to skip over the rest
                # of this
                continue

            using_invalid = self.entity_has_invalid_attributes(entity)
            is_missing = self.entity_missing_attributes(entity)
            cur_guid = entity["guid"]
            if using_invalid:
                report["UsingInvalidAttributes"][cur_guid] = using_invalid

            if is_missing:
                report["MissingRequiredAttributes"][cur_guid] = is_missing

        output = {
            "counts": {k: len(v) for k, v in report.items()},
            "values": report
        }
        output.update({"total": sum(output["counts"].values())})

        return output



import json
from json.decoder import JSONDecodeError
import logging
import re
import requests

from .entity import AtlasClassification, AtlasEntity
from .typedef import BaseTypeDef
from .util import AtlasException, PurviewLimitation, PurviewOnly


class AtlasClient():
    """
    Provides communication between your application and the Apache Atlas
    server with your entities and type definitions.

    :param str endpoint_url:
        The http url for communicating with your Apache Atlas server.
        It will most likely end in /api/atlas/v2.
    :param authentication:
        The method of authentication.
    :type authentication:
        :class:`~pyapacheatlas.auth.base.AtlasAuthBase`
    """

    def __init__(self, endpoint_url, authentication=None):
        super().__init__()
        self.authentication = authentication
        self.endpoint_url = endpoint_url
        self.is_purview = False
        self._purview_url_pattern = r"https:\/\/[a-z0-9-]*?\.(catalog\.purview.azure.com)"
        if re.match(self._purview_url_pattern, self.endpoint_url):
            self.is_purview = True

    def _handle_response(self, resp):
        """
        Safely handle an Atlas Response and return the results if valid.

        :param Response resp: The response from the request method.
        :return: A dict containing the results.
        :rtype: dict
        """

        try:
            results = json.loads(resp.text)
            resp.raise_for_status()
        except JSONDecodeError:
            raise ValueError("Error in parsing: {}".format(resp.text))
        except requests.RequestException as e:
            if "errorCode" in results:
                raise AtlasException(resp.text)
            else:
                raise requests.RequestException(resp.text)

        return results

    def delete_entity(self, guid):
        """
        Delete one or many guids from your Apache Atlas server.

        :param guid: The guid or guids you want to remove.
        :type guid: Union(str,list(str))
        :return:
            An EntityMutationResponse containing guidAssignments,
            mutatedEntities, and partialUpdatedEntities (list).
        :rtype: dict(str, Union(dict,list))
        """
        results = None

        if isinstance(guid, list):
            guid_str = '&guid='.join(guid)
        else:
            guid_str = guid

        atlas_endpoint = self.endpoint_url + \
            "/entity/bulk?guid={}".format(guid_str)
        deleteEntity = requests.delete(
            atlas_endpoint,
            headers=self.authentication.get_authentication_headers())

        results = self._handle_response(deleteEntity)

        return results

    def delete_type(self, name):
        """
        Delete a type based on the given name.

        :param str name: The name of the type you want to remove.
        :return:
            No content, should receive a 204 status code.
        :rtype: None
        """
        results = None

        atlas_endpoint = self.endpoint_url + \
            f"/types/typedef/name/{name}"
        deleteType = requests.delete(
            atlas_endpoint,
            headers=self.authentication.get_authentication_headers())

        try:
            deleteType.raise_for_status()
        except requests.RequestException:
            raise Exception(deleteType.text)

        results = {"message": f"successfully delete {name}"}
        return results

    def get_entity(self, guid=None, qualifiedName=None, typeName=None):
        """
        Retrieve one or many guids from your Atlas backed Data Catalog.

        Returns a dictionary with keys "referredEntities" and "entities". You'll
        want to grab the entities values which is a list of entities.

        You can provide a single guid or a list of guids. You can provide a
        single typeName and multiple qualified names in a list.

        :param guid:
            The guid or guids you want to retrieve. Not used if using typeName
            and qualifiedName.
        :type guid: Union(str, list(str))
        :param qualifiedName:
            The qualified name of the entity you want to find. Must provide
            typeName if using qualifiedName. You may search for multiple
            qualified names under the same type. Ignored if using guid
            parameter.
        :type qualifiedName: Union(str, list(str))
        :param str typeName:
            The type name of the entity you want to find. Must provide
            qualifiedName if using typeName. Ignored if using guid parameter.
        :return:
            An AtlasEntitiesWithExtInfo object which includes a list of
            entities and accessible with the "entities" key.
        :rtype: dict(str, Union(list(dict),dict))
        """
        results = None
        parameters = {}

        if isinstance(guid, list):
            guid_str = '&guid='.join(guid)
        else:
            guid_str = guid

        qualifiedName_params = dict()
        if isinstance(qualifiedName, list):
            qualifiedName_params = {
                f"attr_{idx}:qualifiedName": qname
                for idx, qname in enumerate(qualifiedName)
            }
        else:
            qualifiedName_params = {"attr_0:qualifiedName": qualifiedName}

        if qualifiedName and typeName:
            atlas_endpoint = self.endpoint_url + \
                f"/entity/bulk/uniqueAttribute/type/{typeName}"
            parameters.update(qualifiedName_params)

        else:
            atlas_endpoint = self.endpoint_url + \
                "/entity/bulk?guid={}".format(guid_str)

        getEntity = requests.get(
            atlas_endpoint,
            params=parameters,
            headers=self.authentication.get_authentication_headers()
        )

        results = self._handle_response(getEntity)

        return results

    def partial_update_entity(self, guid=None, typeName=None, qualifiedName=None, attributes={}):
        """
        Partially update an entity without having to construct the entire object
        and its subsequent required attributes. Using guid, you can update a
        single attribute. Using typeName and qualifiedName, you can update
        multiple attributes.

        :param guid:
            The guid for the entity you want to update. Not used if using
            typeName and qualifiedName.
        :type guid: str
        :param qualifiedName:
            The qualified name of the entity you want to update. Must provide
            typeName if using qualifiedName. Ignored if using guid parameter.
        :type qualifiedName: str
        :param str typeName:
            The type name of the entity you want to update. Must provide
            qualifiedName if using typeName. Ignored if using guid parameter.
        :return: The results of your entity update.
        :rtype: dict
        """

        if guid and len(attributes) == 1:
            atlas_endpoint = self.endpoint_url + \
                f"/entity/guid/{guid}"
            attribute_name = list(attributes.keys())[0]
            attribute_value = attributes[attribute_name]
            putEntity = requests.put(
                atlas_endpoint,
                json=attribute_value,
                params={"name": attribute_name},
                headers=self.authentication.get_authentication_headers()
            )
        # TODO: Multiple attributes could be supported for guid by looking up
        # the qualified name and type and then re-running the command with
        # those parameters.
        elif guid:
            raise ValueError(
                "When using guid, attributes can only contain one key and value.")
        elif typeName and qualifiedName:
            atlas_endpoint = self.endpoint_url + \
                f"/entity/uniqueAttribute/type/{typeName}"
            # You have to get the entire existing entity and update its attributes
            get_response = self.get_entity(
                qualifiedName=qualifiedName, typeName=typeName)
            entity = get_response["entities"][0]
            entity["attributes"].update(attributes)
            # Construct it as an AtlasEntityWithInfo
            entityInfo = {"entity": entity,
                          "referredEntities": get_response["referredEntities"]}

            putEntity = requests.put(
                atlas_endpoint,
                json=entityInfo,
                params={"attr:qualifiedName": qualifiedName},
                headers=self.authentication.get_authentication_headers()
            )
        else:
            raise ValueError(
                "The provided combination of arguments is not supported. Either provide a guid or type name and qualified name")

        results = self._handle_response(putEntity)

        return results

    def get_entity_classification(self, guid, classificationName):
        """
        Retrieve a specific entity from the given entity's guid.

        :param str guid: The guid of the entity that you want to query.
        :param str classificationName: The typeName of the classification you
            want to query.
        :return: An AtlasClassification object that contains entityGuid,
            entityStatus, typeName, attributes, and propagate fields.
        :rtype: dict(str, object)
        """
        atlas_endpoint = self.endpoint_url + \
            f"/entity/guid/{guid}/classification/{classificationName}"
        getClassification = requests.get(
            atlas_endpoint,
            headers=self.authentication.get_authentication_headers()
        )
        results = self._handle_response(getClassification)
        return results

    def get_entity_classifications(self, guid):
        """
        Retrieve all classifications from the given entity's guid.

        :param str guid: The entity's guid.
        :return: An AtlasClassifications object that contains keys 'list' (which
            is the list of classifications on the entity), pageSize, sortBy,
            startIndex, and totalCount.
        :rtype: dict(str, object)
        """

        atlas_endpoint = self.endpoint_url + \
            f"/entity/guid/{guid}/classifications"

        getClassification = requests.get(
            atlas_endpoint,
            headers=self.authentication.get_authentication_headers()
        )

        results = self._handle_response(getClassification)

        return results

    def get_entity_header(self, guid=None):
        """
        Retrieve one or many entity headers from your Atlas backed Data Catalog.

        :param guid:
            The guid or guids you want to retrieve. Not used if using typeName
            and qualifiedName.
        :type guid: Union(str, list(str))
        :return:
            An AtlasEntityHeader dict which includes the keys: guid, attributes
            (which is a dict that contains qualifiedName and name keys), an
            array of classifications, and an array of glossary term headers.
        :rtype: dict
        """
        results = None
        parameters = {}

        atlas_endpoint = self.endpoint_url + \
            "/entity/guid/{}/header".format(guid)

        getEntity = requests.get(
            atlas_endpoint,
            params=parameters,
            headers=self.authentication.get_authentication_headers()
        )

        results = self._handle_response(getEntity)

        return results

    def get_relationship(self, guid):
        """
        Retrieve the relationship attribute for the given guid.

        :param str guid: The unique guid for the relationship.
        :return: A dict representing AtlasRelationshipWithExtInfo with the
            relationship (what you probably care about) and referredEntities
            attributes.
        :rtype: dict(str, dict)
        """
        results = None
        atlas_endpoint = self.endpoint_url + f"/relationship/guid/{guid}"

        getResponse = requests.get(
            atlas_endpoint,
            headers=self.authentication.get_authentication_headers()
        )

        results = self._handle_response(getResponse)

        return results

    def get_all_typedefs(self):
        """
        Retrieve all of the type defs available on the Apache Atlas server.

        :return:
            A dict representing an AtlasTypesDef, containing lists of
            type defs wrapped in their corresponding definition types
            {"entityDefs", "relationshipDefs"}.
        :rtype: dict(str, list(dict))
        """
        results = None
        atlas_endpoint = self.endpoint_url + "/types/typedefs"

        getTypeDefs = requests.get(
            atlas_endpoint,
            headers=self.authentication.get_authentication_headers()
        )

        results = self._handle_response(getTypeDefs)

        return results

    def get_typedef(self, type_category=None, guid=None, name=None):
        """
        Retrieve a single type def based on its guid, name, or type category and
        (guid or name). Rule of thumb: Use guid if you have it, use name if you
        want to essentially use duck typing and are testing what keys you're
        reading from the response, or use type_category when you want to
        guarantee the type being returned.

        :param type_category:
            The type category your type def belongs to. You most likely want
            TypeCategory.ENTITY. Optional if name or guid is specified.
        :type type_category:
            :class:`~pyapacheatlas.core.typedef.TypeCategory`
        :param str,optional guid: A valid guid. Optional if name is specified.
        :param str,optional name: A valid name. Optional if guid is specified.
        :return: A dictionary representing an Atlas{TypeCategory}Def.
        :rtype: dict
        """
        results = None
        atlas_endpoint = self.endpoint_url + "/types/"

        # If we are using type category
        if type_category:
            atlas_endpoint = atlas_endpoint + \
                "{}def".format(type_category.value)
        elif guid or name:
            atlas_endpoint = atlas_endpoint + "typedef"
        else:
            raise ValueError(
                "Either guid or name must be defined or type_category and one of guid or name must be defined.")

        if guid:
            atlas_endpoint = atlas_endpoint + '/guid/{}'.format(guid)
        elif name:
            atlas_endpoint = atlas_endpoint + '/name/{}'.format(name)
        else:
            raise ValueError("One of guid or name must be defined.")

        getTypeDef = requests.get(
            atlas_endpoint,
            headers=self.authentication.get_authentication_headers()
        )

        results = self._handle_response(getTypeDef)

        return results

    def get_glossary(self, name="Glossary", guid=None, detailed=False):
        """
        Retrieve the specified glossary by name or guid along with the term
        headers (AtlasRelatedTermHeader: including displayText and termGuid).
        Providing the glossary name only will result in a lookup of all
        glossaries and returns the term headers (accessible via "terms" key)
        for all glossaries.
        Use detailed = True to return the full detail of terms
        (AtlasGlossaryTerm) accessible via "termInfo" key.

        :param str name:
            The name of the glossary to use, defaults to "Glossary". Not
            required if using the guid parameter.
        :param str guid:
            The unique guid of your glossary. Not required if using the
            name parameter.
        :param bool detailed:
            Set to true if you want to pull back all terms and
            not just headers.
        :return:
            The requested glossary with the term headers (AtlasGlossary) or
            with detailed terms (AtlasGlossaryExtInfo).

        :rtype: list(dict)
        """
        results = None

        if guid:
            logging.debug(f"Retreiving a Glossary based on guid: {guid}")
            atlas_endpoint = self.endpoint_url + "/glossary/{}".format(guid)
            if detailed:
                atlas_endpoint = atlas_endpoint + "/detailed"
            getResult = requests.get(
                atlas_endpoint,
                headers=self.authentication.get_authentication_headers()
            )
            results = self._handle_response(getResult)
        else:
            logging.debug(f"Retreiving a Glossary based on name: {name}")
            all_glossaries = self._get_glossaries()
            logging.debug(f"Iterating over {len(all_glossaries)} glossaries")
            for glossary in all_glossaries:
                if glossary["name"] == name:
                    logging.debug(f"Found a glossary named '{name}'")
                    if detailed:
                        logging.debug(
                            f"Recursively calling get_glossary with guid: {glossary['guid']}")
                        results = self.get_glossary(
                            guid=glossary["guid"], detailed=detailed)
                    else:
                        results = glossary
            if results is None:
                raise ValueError(
                    f"Glossary with a name of {name} was not found.")

        return results

    def _get_glossaries(self, limit=-1, offset=0, sort_order="ASC"):
        """
        Retrieve all glossaries and the term headers.

        :param int limit:
            The maximum number of glossaries to pull back.  Does not affect the
            number of term headers included in the results.
        :param int offset: The number of glossaries to skip.
        :param str sort_order: ASC for DESC sort for glossary name.
        :return: The requested glossaries with the term headers.
        :rtype: list(dict)
        """
        results = None
        atlas_endpoint = self.endpoint_url + "/glossary"
        logging.debug("Retreiving all glossaries from catalog")

        # TODO: Implement paging with offset and limit
        getResult = requests.get(
            atlas_endpoint,
            params={"limit": limit, "offset": offset, "sort": sort_order},
            headers=self.authentication.get_authentication_headers()
        )

        results = self._handle_response(getResult)

        return results

    def get_glossary_term(self, guid=None, name=None, glossary_name="Glossary", glossary_guid=None):
        """
        Retrieve a single glossary term based on its guid. Providing only the
        glossary_name will result in a lookup for the glossary guid. If you
        plan on looking up many terms, consider using the get_glossary method
        with the detailed argument set to True. That method will provide all
        glossary terms in a dictionary for faster lookup.

        :param str guid:
            The guid of your term. Not required if name is specified.
        :param str name:
            The name of your term's display text. Overruled if guid is
            provided.
        :param str glossary_name:
            The name of the glossary to use, defaults to "Glossary". Not
            required if using the glossary_guid parameter.
        :param str glossary_guid:
            The unique guid of your glossary. Not required if using the
            glossary_name parameter.
        :return: The requested glossary term as a dict.
        :rtype: dict
        """
        results = None

        if guid is None and name is None:
            raise ValueError("Either guid or name and glossary must be set.")

        if guid:
            atlas_endpoint = self.endpoint_url + \
                "/glossary/term/{}".format(guid)

            getTerms = requests.get(
                atlas_endpoint,
                headers=self.authentication.get_authentication_headers()
            )
            results = self._handle_response(getTerms)
        else:
            terms_in_glossary = self.get_glossary(
                name=glossary_name, guid=glossary_guid)
            for term in terms_in_glossary["terms"]:
                if term["displayText"] == name:
                    _guid = term["termGuid"]
                    results = self.get_glossary_term(guid=_guid)

        return results

    def _get_typedefs_header(self):
        """
        Get the array of AtlasTypeDefHeader that contains category, guid,
        name, and serviceType.  Massage it into a dict based on the available
        categories.

        :return: A dictionary of categories and the names of defined types.
        :rtype: dict(str, list(str))
        """
        atlas_endpoint = self.endpoint_url + "/types/typedefs/headers"
        getHeaders = requests.get(
            atlas_endpoint,
            headers=self.authentication.get_authentication_headers()
        )
        results = self._handle_response(getHeaders)

        output = dict()
        for typedef in results:
            active_category = typedef["category"].lower()+"Defs"
            if active_category not in output:
                output[active_category] = []

            output[active_category].append(typedef["name"])

        return output

    @PurviewLimitation
    def classify_bulk_entities(self, entityGuids, classification):
        """
        Given a single classification, you want to apply it to many entities
        and you know their guid. This call will fail if any one of the guids
        already have the provided classification on that entity.

        :param Union(str,list) entityGuids:
            The guid or guids you want to classify.
        :param classification:
            The AtlasClassification object you want to apply to the entities.
        :type classification:
            Union(dict, :class:`~pyapacheatlas.core.entity.AtlasClassification`)
        :return: A message indicating success. The only key is 'message',
            containing a brief string.
        :rtype: dict(str,Union(list(str),str))
        """
        results = None
        atlas_endpoint = self.endpoint_url + "/entity/bulk/classification"

        if isinstance(classification, AtlasClassification):
            classification = classification.to_json()

        classification_name = classification["typeName"]

        if isinstance(entityGuids, str):
            entityGuids = [entityGuids]
        elif isinstance(entityGuids, list):
            pass
        else:
            raise TypeError(
                "guid should be str or list, not {}".format(type(entityGuids)))

        payload = {
            # TODO: Accept AtlasClassification class
            "classification": classification,
            "entityGuids": entityGuids
        }

        postBulkClassifications = requests.post(
            atlas_endpoint,
            json=payload,
            headers=self.authentication.get_authentication_headers()
        )

        try:
            postBulkClassifications.raise_for_status()
        except requests.RequestException:
            raise AtlasException(postBulkClassifications.text)

        results = {"message": f"Successfully assigned {classification_name}",
                   "entityGuids": entityGuids
                   }
        return results

    def _classify_entity_adds(self, guid, classifications):
        """
        Update a given entity guid with the provided classifications.

        :param str guid: The guid you want to classify.
        :param list(dict) classifications:
            The list of AtlasClassification object you want to apply to the
            entity.
        :return: The name of the classification provided are returned.
        :rtype: list(str)
        """
        results = None
        atlas_endpoint = self.endpoint_url + \
            f"/entity/guid/{guid}/classifications"

        postAddMultiClassifications = requests.post(
            atlas_endpoint,
            json=classifications,
            headers=self.authentication.get_authentication_headers()
        )

        try:
            postAddMultiClassifications.raise_for_status()
        except requests.RequestException:
            raise Exception(postAddMultiClassifications.text)

        results = [c["typeName"] for c in classifications]
        return results

    def _classify_entity_updates(self, guid, classifications):
        """
        Update a given entity guid with the provided classifications.

        :param str guid: The guid you want to classify.
        :param list(dict) classifications:
            The list of AtlasClassification object you want to update to the
            entity.
        :return: The name of the classification provided are returned.
        :rtype: list(str)
        """
        results = None
        atlas_endpoint = self.endpoint_url + \
            f"/entity/guid/{guid}/classifications"

        putUpdateMultiClassifications = requests.put(
            atlas_endpoint,
            json=classifications,
            headers=self.authentication.get_authentication_headers()
        )

        try:
            putUpdateMultiClassifications.raise_for_status()
        except requests.RequestException:
            raise Exception(putUpdateMultiClassifications.text)

        results = [c["typeName"] for c in classifications]
        return results

    @PurviewLimitation
    def classify_entity(self, guid, classifications, force_update=False):
        """
        Given a single entity, you want to apply many classifications.
        This call will fail if any one of the classifications exist on the
        entity already, unless you choose force_update=True.

        force_update will query the existing entity and sort the classifications
        into NEW (post) and UPDATE (put) and do two requests to add and update.

        force_update is not transactional, it performs adds first. If the add
        succeeds, it moves on to the updates. If the update fails, the adds
        will continue to exist on the Atlas server and will not be rolledback.

        An error can occur if, for example, the classification has some required
        attribute that you do not provide.

        :param str guid: The guid you want to classify.
        :param classifications:
            The list of AtlasClassification object you want to apply to the
            entities.
        :type classification: 
            Union(dict, :class:`~pyapacheatlas.core.entity.AtlasClassification`)
        :param bool force_update: Mark as True if any of your classifications
            may already exist on the given entity.
        :return: A message indicating success and which classifications were
            'updates' vs 'adds' for the given guid.
        :rtype: dict(str, str)
        """
        results = None
        adds = []
        updates = []

        if isinstance(classifications, dict):
            classifications = [classifications]
        elif isinstance(classifications, AtlasClassification):
            classifications = [classifications.to_json()]
        elif isinstance(classifications, list):
            classifications = [
                c.to_json()
                if isinstance(c, AtlasClassification)
                else c for c in classifications]
        else:
            raise TypeError("classifications should be dict or list, not {}".format(
                type(classifications)))

        if force_update:
            # Get the existing entity's classifications
            existing_classifications = set([
                c["typeName"] for c in
                self.get_entity_classifications(guid=guid)["list"]
            ])

            # Sort the list into adds and updates (if exists)
            temp_adds = []
            temp_updates = []
            for classification in classifications:
                if classification["typeName"] in existing_classifications:
                    temp_updates.append(classification)
                else:
                    temp_adds.append(classification)

            # execute adds
            if len(temp_adds) > 0:
                adds = self._classify_entity_adds(guid, temp_adds)
            # execute updates
            if len(temp_updates) > 0:
                updates = self._classify_entity_updates(guid, temp_updates)
        else:
            # Assuming this is all new adds
            # execute adds
            adds = self._classify_entity_adds(guid, classifications)

        results = {
            "message": "Successfully assigned classifications",
            "guid": guid,
            "adds": adds,
            "updates": updates
        }
        return results

    def declassify_entity(self, guid, classificationName):
        """
        Given an entity guid and a classification name, remove the
        classification from the given entity.

        :param str guid: The guid for the entity that needs to be updated.
        :param str classificationName:
            The name of the classification to be deleted.
        :return: A success message repeating what was deleted. The only key
            is 'message', containing the classification name and guid.
        :rtype: dict(str, str)
        """
        results = None
        atlas_endpoint = self.endpoint_url + \
            f"/entity/guid/{guid}/classification/{classificationName}"

        deleteEntityClassification = requests.delete(
            atlas_endpoint,
            headers=self.authentication.get_authentication_headers()
        )

        try:
            deleteEntityClassification.raise_for_status()
        except requests.RequestException:
            raise Exception(deleteEntityClassification.text)

        results = {"message":
                   f"Successfully removed classification: {classificationName} from {guid}.",
                   "guid": guid,
                   }
        return results

    @staticmethod
    def _prepare_type_upload(typedefs=None, **kwargs):
        """
        Massage the type upload. See rules in upload_typedefs.
        """
        payload = {}
        required_keys = ["classificationDefs", "entityDefs",
                         "enumDefs", "relationshipDefs", "structDefs"]

        # If typedefs is defined as a dict and it contains at least one of the
        # required keys for the TypeREST definition.
        if isinstance(typedefs, dict) and len(set(typedefs.keys()).intersection(required_keys)) > 0:
            payload = typedefs
        # It isn't in the standard form but is it defined?
        elif typedefs is not None:
            # Assuming this is a single typedef
            key = None
            if isinstance(typedefs, BaseTypeDef):
                key = typedefs.category.lower() + "Defs"
                val = [typedefs.to_json()]
            elif isinstance(typedefs, dict):
                key = typedefs["category"].lower() + "Defs"
                val = [typedefs]
            else:
                raise NotImplementedError(
                    "Uploading an object of type '{}' is not supported."
                    .format(type(typedefs))
                )
            payload = {key: val}
        # Did we set any of the xDefs as arguments?
        elif len(set(kwargs.keys()).intersection(required_keys)) > 0:
            for typeRestKey in required_keys:
                # Did we specify this key?
                if typeRestKey in kwargs.keys():
                    payload[typeRestKey] = [
                        t.to_json() if isinstance(t, BaseTypeDef) else t
                        for t in kwargs[typeRestKey]
                    ]
        else:
            raise RuntimeError(
                f"Failed to upload typedefs for arguments: {kwargs}"
            )
        return payload

    def upload_typedefs(self, typedefs=None, force_update=False, **kwargs):
        """
        Provides a way to upload a single or multiple type definitions.
        If you provide one type def, it will format the required wrapper
        for you based on the type category.

        If you want to upload multiple type defs or typedefs of different
        category, you can pass the in kwargs `entityDefs`, `classificationDefs`,
        `enumDefs`, `relationshipDefs`, `structDefs` which take in a list of
        dicts or appropriate TypeDef objects.

        Otherwise, you can pass in the wrapper yourself (e.g. {"entityDefs":[],
        "relationshipDefs":[]}) by providing that dict to the typedefs
        parameter. If the dict you pass in contains at least one of these Def
        fields it will be considered valid and an upload will be attempted.

        typedefs also takes in a BaseTypeDef object or a valid AtlasTypeDef
        json / dict. If you provide a value in typedefs, it will ignore the
        kwargs parameters.

        When using force_update, it will look up all existing types and see
        if any of your provided types exist.  If they do exist, they will be
        updated. If they do not exist, they will be issued as new. New types
        are uploaded first. Existing types are updated second. There are no
        transactional updates.  New types can succeed and be inserted while
        a batch of existing types can fail and not be updated.

        :param typedefs: The set of type definitions you want to upload.
        :type typedefs: Union(dict, :class:`~pyapacheatlas.core.typedef.BaseTypeDef`)
        :param bool force_update:
            Set to True if your typedefs contains any existing entities.
        :return: The results of your upload attempt from the Atlas server.
            :rtype: dict

        Kwargs:
            :param entityDefs: EntityDefs to upload.
            :type entityDefs: list( Union(:class:`~pyapacheatlas.core.typedef.BaseTypeDef`, dict))
            :param classificationDefs: classificationDefs to upload.
            :type classificationDefs: list( Union(:class:`~pyapacheatlas.core.typedef.BaseTypeDef`, dict))
            :param enumDefs: enumDefs to upload.
            :type enumDefs: list( Union(:class:`~pyapacheatlas.core.typedef.BaseTypeDef`, dict))
            :param relationshipDefs: relationshipDefs to upload.
            :type relationshipDefs: list( Union(:class:`~pyapacheatlas.core.typedef.BaseTypeDef`, dict))
            :param structDefs: structDefs to upload.
            :type structDefs: list( Union(:class:`~pyapacheatlas.core.typedef.BaseTypeDef`, dict))

        Returns:

        """
        # Should this take a list of type defs and figure out the formatting
        # by itself?
        # Should you pass in a AtlasTypesDef object and be forced to build
        # it yourself?
        results = None
        atlas_endpoint = self.endpoint_url + "/types/typedefs"

        payload = AtlasClient._prepare_type_upload(typedefs, **kwargs)

        if not force_update:
            # This is just a plain push of new entities
            upload_typedefs_results = requests.post(
                atlas_endpoint, json=payload,
                headers=self.authentication.get_authentication_headers()
            )
            results = self._handle_response(upload_typedefs_results)
        else:
            # Look up all entities by their header
            types_from_client = self._get_typedefs_header()
            existing_types = dict()
            new_types = dict()

            # Loop over the intended upload and see if they exist already
            # If they do not exist, shuffle them to the new_types upload.
            # if they do exist, shuffle to the existing types upload.
            for cat, typelist in payload.items():
                existing_types[cat] = []
                new_types[cat] = []
                for t in typelist:
                    if t["name"] in types_from_client[cat]:
                        existing_types[cat].append(t)
                    else:
                        new_types[cat].append(t)

            upload_new = requests.post(
                atlas_endpoint, json=new_types,
                headers=self.authentication.get_authentication_headers()
            )
            results_new = self._handle_response(upload_new)

            upload_exist = requests.put(
                atlas_endpoint, json=existing_types,
                headers=self.authentication.get_authentication_headers()
            )
            results_exist = self._handle_response(upload_exist)

            # Merge the results
            results = results_new
            for cat, updatedtypelist in results_exist.items():
                if cat not in results:
                    results[cat] = []
                results[cat].extend(updatedtypelist)

        return results

    @staticmethod
    def _prepare_entity_upload(batch):
        """
        Massages the batch to be in the right format and coerces to json/dict.
        Supports list of dicts, dict of single entity, dict of AtlasEntitiesWithExtInfo.

        :param batch: The batch of entities you want to upload.
        :type batch: Union(list(dict), dict))
        :return: Provides a dict formatted in the Atlas entity bulk upload.
        :rtype: dict(str, list(dict))
        """
        payload = batch
        required_keys = ["entities"]

        if isinstance(batch, list):
            # It's a list, so we're assuming it's a list of entities
            # Handles any type of AtlasEntity and mixed batches of dicts
            # and AtlasEntities
            dict_batch = [e.to_json() if isinstance(
                e, AtlasEntity) else e for e in batch]
            payload = {"entities": dict_batch}
        elif isinstance(batch, dict):
            current_keys = list(batch.keys())

            # Does the dict entity conform to the required pattern?
            if not any([req in current_keys for req in required_keys]):
                # Assuming this is a single entity
                # DESIGN DECISION: I'm assuming, if you're passing in
                # json, you know the schema and I will not support
                # AtlasEntity here.
                payload = {"entities": [batch]}
        elif isinstance(batch, AtlasEntity):
            payload = {"entities": [batch.to_json()]}
        else:
            raise NotImplementedError(
                f"Uploading type: {type(batch)} is not supported.")

        return payload

    def upload_entities(self, batch):
        """
        Upload entities to your Atlas backed Data Catalog.

        :param batch:
            The batch of entities you want to upload. Supports a single dict,
            AtlasEntity, list of dicts, list of atlas entities.
        :type batch:
            Union(dict, :class:`~pyapacheatlas.core.entity.AtlasEntity`,
            list(dict), list(:class:`~pyapacheatlas.core.entity.AtlasEntity`) )
        :return: The results of your bulk entity upload.
        :rtype: dict
        """
        # TODO Include a Do Not Overwrite call
        results = None
        atlas_endpoint = self.endpoint_url + "/entity/bulk"

        payload = AtlasClient._prepare_entity_upload(batch)

        postBulkEntities = requests.post(
            atlas_endpoint,
            json=payload,
            headers=self.authentication.get_authentication_headers()
        )

        results = self._handle_response(postBulkEntities)

        return results

    def upload_relationship(self, relationship):
        """
        Upload a AtlasRelationship json. Should take the form of the following::

            {
                "typeName": "hive_table_columns",
                "attributes": {},
                "guid": -100,
                "end1": {
                    "guid": assignments["-1"]
                },
                "end2": {
                    "guid": assignments["-5"]
                    }
            }

        :param dict relationship: The relationship you want to upload.
        :return: The results of your relationship upload.
        :rtype: dict
        """
        # TODO Include a Do Not Overwrite call
        results = None
        atlas_endpoint = self.endpoint_url + "/relationship"

        # TODO: Handling Updates instead of just creates
        relationshipResp = requests.post(
            atlas_endpoint,
            json=relationship,
            headers=self.authentication.get_authentication_headers()
        )

        results = self._handle_response(relationshipResp)

        return results

    def upload_terms(self, batch, force_update=False):
        """
        Upload terms to your Atlas backed Data Catalog. Supports Purview Term
        Templates by passing in an attributes field with the term template's
        name as a field within attributes and an object of the required and
        optional fields.

        :param batch: A list of AtlasGlossaryTerm objects to be uploaded.
        :type batch: list(dict)
        :return:
            A list of AtlasGlossaryTerm objects that are the results from
            your upload.
        :rtype: list(dict)
        """
        # TODO Include a Do Not Overwrite call
        results = None
        atlas_endpoint = self.endpoint_url + "/glossary/terms"

        postResp = requests.post(
            atlas_endpoint,
            json=batch,
            headers=self.authentication.get_authentication_headers()
        )

        results = self._handle_response(postResp)

        return results

    def _search_generator(self, search_params, starting_offset=0):
        """
        Generator to page through the search query results.
        """
        atlas_endpoint = self.endpoint_url + "/search/advanced"
        offset = starting_offset

        while True:
            postSearchResults = requests.post(
                atlas_endpoint,
                json=search_params,
                headers=self.authentication.get_authentication_headers()
            )
            results = self._handle_response(postSearchResults)
            return_values = results["value"]
            return_count = len(return_values)

            if return_count == 0:
                return

            offset = offset + return_count
            search_params["offset"] = offset

            for sub_result in return_values:
                try:
                    yield sub_result
                except StopIteration:
                    return

    @PurviewOnly
    def search_entities(self, query, limit=50, search_filter=None, starting_offset=0):
        """
        Search entities based on a query and automaticall handles limits and
        offsets to page through results.

        The limit provides how many records are returned in each batch with a
        maximum of 1,000 entries per page.

        :param str query: The search query to be executed.
        :param int limit:
            A non-zero integer representing how many entities to
            return for each page of the search results.
        :param dict search_filter: A search filter to reduce your results.
        :return: The results of your search as a generator.
        :rtype: Iterator(dict)
        """

        if limit > 1000 or limit < 1:
            raise ValueError(
                "The limit parameter must be non-zero and less than 1,000."
            )

        search_params = {
            "keywords": query,
            "limit": limit,
            "offset": 0
        }
        # TODO: Make this smarter, make it easier to create filters
        # without having to know how to make a filter object.
        if search_filter:
            # Example search filter might look like:
            # {"filter": {"typeName": "DataSet", "includeSubTypes": True} }
            search_params.update({"filter": search_filter})

        search_generator = self._search_generator(
            search_params, starting_offset=starting_offset)

        return search_generator

    def get_entity_lineage(self, guid, depth=3, width=10, direction="BOTH", includeParent=False, getDerivedLineage=False):
        """
        Gets lineage info about the specified entity by guid.

        :param str guid: The guid of the entity for which you want to
            retrieve lineage.
        :param int depth: The number of hops for lineage
        :param int width: The number of max expanding width in lineage
        :param str direction: The direction of the lineage, which could
            be INPUT, OUTPUT or BOTH.
        :param bool includeParent: True to include the parent chain in
            the response
        :param bool getDerivedLineage: True to include derived lineage in
            the response
        :return: A dict representing AtlasLineageInfo with an array
            of parentRelations and an array of relations
        :rtype: dict(str, dict)
        """
        direction = direction.strip().upper()
        assert direction in (
            "BOTH", "INPUT", "OUTPUT"), "Invalid direction '{}'.  Valid options are: BOTH, INPUT, OUTPUT".format(direction)

        atlas_endpoint = self.endpoint_url + \
            f"/lineage/{guid}"

        getLineageRequest = requests.get(
            atlas_endpoint,
            params={"depth": depth, "width": width, "direction": direction,
                    "includeParent": includeParent, "getDerivedLineage": getDerivedLineage},
            headers=self.authentication.get_authentication_headers()
        )
        results = self._handle_response(getLineageRequest)
        return results


class PurviewClient(AtlasClient):
    """
    Provides communication between your application and the Azure Purview
    service. Simplifies the requirements for knowing the endpoint url and
    requires only the Purview account name.

    :param str account_name:
        Your Purview account name.
    :param authentication:
        The method of authentication.
    :type authentication:
        :class:`~pyapacheatlas.auth.base.AtlasAuthBase`
    """

    def __init__(self, account_name, authentication=None):
        endpoint_url = f"https://{account_name.lower()}.catalog.purview.azure.com/api/atlas/v2"
        super().__init__(endpoint_url, authentication)

    @PurviewOnly
    def get_entity_next_lineage(self, guid, direction, getDerivedLineage=False, offset=0, limit=-1):
        """
        Returns immediate next level lineage info about entity with pagination

        :param str guid: The guid of the entity for which you want to
            retrieve lineage.
        :param str direction: The direction of the lineage, which could
            be INPUT or OUTPUT.
        :param bool getDerivedLineage: True to include derived lineage in
            the response
        :param int offset: The offset for pagination purpose.
        :param int limit: The page size - by default there is no paging.
        :return: A dict representing AtlasLineageInfo with an array
            of parentRelations and an array of relations
        :rtype: dict(str, dict)
        """
        direction = direction.strip().upper()
        assert direction in (
            "INPUT", "OUTPUT"), "Invalid direction '{}'.  Valid options are: INPUT, OUTPUT".format(direction)

        atlas_endpoint = self.endpoint_url + \
            f"/lineage/{guid}/next"

        # TODO: Implement paging with offset and limit
        getLineageRequest = requests.get(
            atlas_endpoint,
            params={"direction": direction, "getDerivedLineage": getDerivedLineage,
                    "offset": offset, "limit": limit},
            headers=self.authentication.get_authentication_headers()
        )
        results = self._handle_response(getLineageRequest)
        return results

    def import_terms(self, csv_path, glossary_name="Glossary", glossary_guid=None):
        """
        Bulk import terms from an existing csv file. If you are using the system
        default, you must include the following headers:
        Name,Definition,Status,Related Terms,Synonyms,Acronym,Experts,Stewards

        For custom term templates, additional attributes must include
        [Attribute][termTemplateName]attributeName as the header.

        :param str csv_path: Path to CSV that will be imported.
        :param str glossary_name:
            Name of the glossary. Defaults to 'Glossary'. Not used if
            glossary_guid is provided.
        :param str glossary_guid:
            Guid of the glossary, optional if glossary_name is provided.
            Otherwise, this parameter takes priority over glossary_name.

        :return:
            A dict that contains an `id` that you can use in
            `import_terms_status` to get the status of the import operation.
        :rtype: dict
        """
        results = None
        if glossary_guid:
            atlas_endpoint = self.endpoint_url + \
                f"/glossary/{glossary_guid}/terms/import"
        elif glossary_name:
            atlas_endpoint = self.endpoint_url + \
                f"/glossary/name/{glossary_name}/terms/import"
        else:
            raise ValueError(
                "Either glossary_name or glossary_guid must be defined.")

        headers = self.authentication.get_authentication_headers()
        # Pop the default of application/json so that request can fill in the
        # multipart/form-data; boundary=xxxx that is automatically generated
        # when using the files argument.
        headers.pop("Content-Type")

        postResp = requests.post(
            atlas_endpoint,
            files={'file': ("file", open(csv_path, 'rb'))},
            headers=headers
        )

        results = self._handle_response(postResp)

        return results

    @PurviewOnly
    def import_terms_status(self, operation_guid):
        """
        Get the operation status of a glossary term import activity. You get
        the operation_guid after executing the `import_terms` method and find
        the `id` field in the response dict/json.

        :param str operation_guid: The id of the import operation.
        :return: The status of the import operation as a dict. The dict includes
            a field called `status` that will report back RUNNING, SUCCESS, or
            FAILED. Other fields include the number of terms detected and
            number of errors.
        :rtype: dict
        """
        results = None
        atlas_endpoint = self.endpoint_url + \
            f"/glossary/terms/import/{operation_guid}"

        postResp = requests.get(
            atlas_endpoint,
            headers=self.authentication.get_authentication_headers()
        )

        results = self._handle_response(postResp)

        return results

    @PurviewOnly
    def export_terms(self, guids, csv_path, glossary_name="Glossary", glossary_guid=None):
        """
        :param list(str) guids: List of guids that should be exported as csv.
        :param str csv_path: Path to CSV that will be imported.
        :param str glossary_name:
            Name of the glossary. Defaults to 'Glossary'. Not used if
            glossary_guid is provided.
        :param str glossary_guid:
            Guid of the glossary, optional if glossary_name is provided.
            Otherwise, this parameter takes priority over glossary_name.
            Providing glossary_guid is also faster as you avoid a lookup based
            on glossary_name.

        :return: A csv file is written to the csv_path.
        :rtype: None
        """
        if glossary_guid:
            # Glossary guid is defined so we don't need to look up the guid
            pass
        elif glossary_name:
            glossary = self.get_glossary(glossary_name)
            glossary_guid = glossary["guid"]
        else:
            raise ValueError(
                "Either glossary_name or glossary_guid must be defined.")

        results = None
        atlas_endpoint = self.endpoint_url + \
            f"/glossary/{glossary_guid}/terms/export"

        postResp = requests.post(
            atlas_endpoint,
            json=guids,
            headers=self.authentication.get_authentication_headers()
        )

        # Can't use handle response since it expects json
        try:
            postResp.raise_for_status()
        except requests.RequestException as e:
            if "errorCode" in postResp:
                raise AtlasException(postResp.text)
            else:
                raise requests.RequestException(postResp.text)

        with open(csv_path, 'wb') as fp:
            fp.write(postResp.content)

        return None



from .client import AtlasClient, PurviewClient
from .entity import AtlasClassification, AtlasEntity, AtlasProcess
from .typedef import (
    AtlasAttributeDef,
    ClassificationTypeDef,
    EntityTypeDef,
    RelationshipTypeDef,
    TypeCategory
)
from .util import AtlasException



from ..core import typedef


def to_entity_def(entity_name, attributes, super_types=["DataSet"]):
    """
    :param str entity_name: The name of the entity definition to create.
    :param list(str) attributes:
        The attributes that should be converted into an
        entity def list of attributes.
    :param list(str) super_types: The super types to inherit from.

    :return: An entity definition with the provided attributes and the default
        settings of attributes as defined in
        :class:`~pyapacheatlas.core.typedef.AtlasAttributeDef`.
    :rtype: dict
    """
    # Take in a list of attribute names
    # Special columns: name, superTypes
    # TODO: Handling parent types already in place?
    # For each attribute, Create a default attribute def
    # Add it to the Entity Def

    attributeDefs = [
        typedef.AtlasAttributeDef(name=i).to_json() for i in attributes]

    entityDef = typedef.EntityTypeDef(
        name=entity_name,
        superTypes=super_types,
        attributeDefs=attributeDefs
    )

    return entityDef.to_json()



from collections import OrderedDict
import json
import warnings

from ..core import AtlasEntity, AtlasProcess
from . import util as reader_util


class LineageMixIn():
    """
    A MixIn to support the :class:`~pyapacheatlas.readers.reader.Reader` with
    the table and column level lineages.
    """

    def _update_entity_and_array(self, entity, mutableOutput):
        """
        Take in an AtlasEntity and list of AtlasEntities.  If the entity
        exists in the mutableOutput, remove the entity from the list and
        merge the incoming entity with the popped entity.  Append that
        merged entity onto the mutableOutput.

        :param entity:
            The entity to look up in the mutableOutput.
        :param mutableOutput:
            A list of Atlas Entities to search through and update if
            `entity` is found.
        :type entity:
            :class:`~pyapacheatlas.core.entity.AtlasEntity`
        :type mutableOutput:
            list(:class:`~pyapacheatlas.core.entity.AtlasEntity`)
        """
        if entity in mutableOutput:
            # Assumes things like name, type name, are consistent
            poppable_index = mutableOutput.index(entity)
            popped_target = mutableOutput.pop(poppable_index)
            # Update the newly created entity with the previous versions
            # attributes
            entity.merge(popped_target)
        mutableOutput.append(entity)

    def _add_table_lineage_entity(self, row, header, attributes, mutableOutput):
        """
        Create an AtlasEntity object and add it to the mutableOutput with the
        desired attributes.
        """
        entity = AtlasEntity(
            name=row[header["table"]],
            typeName=row[header["type"]],
            # qualifiedName can be overwritten via the attributes
            # functionality
            qualified_name=row[header["table"]],
            guid=self.guidTracker.get_guid(),
            attributes=attributes,
            classifications=reader_util.string_to_classification(
                row.get(header["classifications"]),
                self.config.value_separator
            )
        )

        self._update_entity_and_array(entity, mutableOutput)

        return entity

    def parse_table_lineage(self, json_rows):
        """
        Converts the "tables" information into Atlas Entities for Target,
        Source, and Process types.  Currently only support one target from
        one source.

        :param json_rows:
            A list of dicts that contain the converted tables of your column
            spreadsheet.
        :type json_rows: list(dict(str,str))
        :return:
            A list of atlas entities that represent your source, target,
            and table processes.
        :rtype: list(:class:`~pyapacheatlas.core.entity.AtlasEntity`)

        """
        # Required attributes
        # NOTE: Classification is not actually required but it's being
        # included to avoid being roped in as an attribute
        _required_headers = ["table", "type", "classifications"]
        source_header = {k: "{} {}".format(
            self.config.source_prefix, k) for k in _required_headers}
        target_header = {k: "{} {}".format(
            self.config.target_prefix, k) for k in _required_headers}
        process_header = {k: "{} {}".format(
            self.config.process_prefix, k) for k in ["name", "type"]}

        # Read in all Source and Target entities
        output = list()  # TODO: Change to a dict to facilitate lookups
        for row in json_rows:
            # Set up defaults
            target_entity, source_entity, process_entity = None, None, None

            # Always expecting a TARGET in the sheet
            target_attributes = reader_util.columns_matching_pattern(
                row, self.config.target_prefix,
                does_not_match=list(target_header.values())
            )
            target_entity = self._add_table_lineage_entity(
                row, target_header, target_attributes, output
            )

            if row[source_header["table"]] is not None:
                # There is a source table
                source_attributes = reader_util.columns_matching_pattern(
                    row, self.config.source_prefix,
                    does_not_match=list(source_header.values())
                )
                source_entity = self._add_table_lineage_entity(
                    row, source_header, source_attributes, output
                )

            # Map the source and target tables to a process
            if row[process_header["name"]] is not None:
                # There is a process
                inputs_to_process = [] if source_entity is None else [
                    source_entity.to_json(minimum=True)]
                process_attributes = reader_util.columns_matching_pattern(
                    row, self.config.process_prefix,
                    does_not_match=list(process_header.values())
                )

                process_entity = AtlasProcess(
                    name=row[process_header["name"]],
                    typeName=row[process_header["type"]],
                    qualified_name=row[process_header["name"]],
                    guid=self.guidTracker.get_guid(),
                    inputs=inputs_to_process,
                    outputs=[target_entity.to_json(minimum=True)],
                    attributes=process_attributes
                )

                self._update_entity_and_array(process_entity, output)

        # Return all entities
        return output

    def _update_parent_table_cache(self, parent_table_name, known_tables, atlas_entities):
        """
        Updates the known_tables variable to include the parent table entity if
        it does not exist already in known_tables.  Looks up the table in the
        atlas_entities if it does not exist.
        :param str parent_table_name:
            The parent table name (not qualified name)
        :param dict(str, dict): All tables we've seen so far.
        :param atlas_entities:
            A list of :class:`~pyapacheatlas.core.entity.AtlasEntity`
            containing the referred entities.
        :type atlas_entities:
            list(:class:`~pyapacheatlas.core.entity.AtlasEntity`)
        :return: None
        :rtype: NoneType
        """
        if parent_table_name not in known_tables:
            parent_entity = reader_util.first_entity_matching_attribute(
                "name", parent_table_name, atlas_entities)
            known_tables[parent_table_name] = parent_entity

    def _find_column_type(self, parentTypeName, atlas_typedefs):
        """
        For the given type, find a relationship def that includes "columns" in
        the end def1 (presumably the table end of the relationship) and then
        extract the type from the end def2 (presumably the column end of the)
        relationship.

        :param str parentTypeName: The name of the table type.
        :param dict(str, dict) atlas_typedefs:
            The results of requesting all type defs from Apache Atlas,
            including entityDefs, relationshipDefs, etc.  relationshipDefs
            are the only values used.
        """
        columns_relationship = reader_util.first_relationship_that_matches(
            end_def="endDef1",
            end_def_type=parentTypeName,
            end_def_name="columns",
            relationship_typedefs=atlas_typedefs["relationshipDefs"]
        )
        column_type = columns_relationship["endDef2"]["type"]
        return column_type

    def _insert_column_entity(self, prefix, row, column_type, headers, column_entities, parent_table_entity):
        """

        :param str prefix: The prefix used to filter down the attributes.
        :param dict(str, str) row:
            Represents a row containing the attributes to convert into an atlas
            entity.
        :param str column_type: The column type to be used for this entity.
        :param dict(str, str) headers:
            The headers to aide in looking up field values.
        :param column_entities:
            The column entities dictionary that will be inserted into.
        :type column_entities:
            list(`:class:~pyapacheatlas.core.entity.AtlasEntity`)
        :param parent_table_entity: The parent table entity for this column.
        :type parent_table_entity:
            `:class:~pyapacheatlas.core.entity.AtlasEntity`
        """
        _qual_name = reader_util._make_col_qual_name(
            row[headers["column"]], parent_table_entity.name)

        _clean_attribs = reader_util.columns_matching_pattern(
            row, prefix,
            does_not_match=list(headers.values())
        )
        _attributes = self._organize_attributes(
            _clean_attribs,
            column_entities,
            []
        )
        _attributes["relationshipAttributes"].update(
            {
                "table": parent_table_entity.to_json(
                    minimum=True)
            }
        )
        _classifications = reader_util.string_to_classification(
            row.get(headers["classifications"])
        )

        # There should always be a target
        new_entity = AtlasEntity(
            name=row[headers["column"]],
            typeName=column_type,
            # qualifiedName can be overwritten via the attributes
            # functionality
            qualified_name=_qual_name,
            guid=self.guidTracker.get_guid(),
            attributes=_attributes["attributes"],
            # TODO: Make the relationship name dynamic instead of only table
            relationshipAttributes=_attributes["relationshipAttributes"],
            classifications=_classifications
        )
        if new_entity in column_entities:
            popped_entity = column_entities.pop(new_entity)
            new_entity.merge(popped_entity)
        # Add to outputs
        payload = {new_entity.qualifiedName: new_entity}
        column_entities.update(payload)

        return _qual_name

    def parse_column_lineage(self, json_rows, atlas_entities, atlas_typedefs, use_column_mapping=False):
        """
        :param json_rows:
            A list of dicts that contain the converted rows of your column
            spreadsheet.
        :type json_rows: list(dict(str,str))
        :param atlas_entities:
            A list of :class:`~pyapacheatlas.core.entity.AtlasEntity`
            containing the referred entities.
        :type atlas_entities:
            list(:class:`~pyapacheatlas.core.entity.AtlasEntity`)
        :param dict(str,list(dict)) atlas_typedefs:
            The results of requesting all type defs from Apache Atlas,
            including entityDefs, relationshipDefs, etc.  relationshipDefs
            are the only values used.
        :param bool use_column_mapping:
            Should the table processes include the columnMappings attribute
            that represents Column Lineage in Azure Data Catalog.
            Defaults to False.
        :return:
            A list of atlas entities that represent your column source, target,
            and column lineage processes.
        :rtype: list(:class:`~pyapacheatlas.core.entity.AtlasEntity`)

        """
        # Required attributes
        # NOTE: Classification is not actually required but it's being
        # included to avoid being roped in as an attribute
        _required_headers = ["table", "column", "classifications"]
        source_header = {k: "{} {}".format(
            self.config.source_prefix, k) for k in _required_headers}
        target_header = {k: "{} {}".format(
            self.config.target_prefix, k) for k in _required_headers}

        transformation_column_header = self.config.column_transformation_name
        # No required process headers

        columnEntitiesOutput = OrderedDict()
        # `tables` Stores all of the table entities by their name seen in the
        # loop for faster lookup
        tables = {}
        # dataset_mapping:
        # {"input_table":"","output_table":"","columnMapping":[]}
        dataset_mapping = {}
        # Caches all of the table processes seen in the loop for faster lookup
        table_and_proc_mappings = {}

        for row in json_rows:
            # Set up defaults
            target_qual_name, source_qual_name, process_qual_name = None, None, None
            table_process = None
            # Given the existing table entity in atlas_entities,
            # look up the appropriate column type

            self._update_parent_table_cache(
                row[target_header["table"]], tables, atlas_entities)

            target_parent = tables[row[target_header["table"]]]
            target_column_type = self._find_column_type(
                target_parent.typeName, atlas_typedefs)

            target_qual_name = self._insert_column_entity(
                self.config.target_prefix, row, target_column_type,
                target_header,
                columnEntitiesOutput, target_parent
            )

            # Source Column is optiona in the spreadsheet
            if row[source_header["table"]] is not None:
                # Given the existing source table entity in atlas_entities,
                # look up the appropriate column type
                self._update_parent_table_cache(
                    row[source_header["table"]], tables, atlas_entities)

                source_parent = tables[row[source_header["table"]]]
                source_column_type = self._find_column_type(
                    source_parent.typeName, atlas_typedefs)

                source_qual_name = self._insert_column_entity(
                    self.config.source_prefix, row, source_column_type,
                    source_header,
                    columnEntitiesOutput, source_parent
                )
                table_process = reader_util.first_process_containing_io(
                    row[source_header["table"]], row[target_header["table"]],
                    atlas_entities)

            # Given the existing process that with target table and source
            # table types, look up the appropriate column_lineage type
            # LIMITATION: Prevents you from specifying multiple processes
            # for the same input and output tables
            if row[source_header["table"]] is None:
                table_process = reader_util.first_process_containing_io(
                    "*", row[target_header["table"]], atlas_entities)

            if table_process.name in table_and_proc_mappings:
                process_type = table_and_proc_mappings[table_process.name]["column_lineage_type"]
            else:
                process_type = reader_util.from_process_lookup_col_lineage(
                    table_process.name,
                    atlas_entities,
                    atlas_typedefs["relationshipDefs"]
                )
                table_and_proc_mappings[table_process.name] = {
                    "column_lineage_type": process_type
                }

            # Assuming there is always a Process for adding at least the
            # target table
            process_attributes = reader_util.columns_matching_pattern(
                row, self.config.process_prefix)
            process_attributes.update({"dependencyType": "SIMPLE"})
            if row[transformation_column_header] is not None:
                process_attributes.update(
                    {"dependencyType": "EXPRESSION",
                     "expression": row[transformation_column_header]
                     })

            process_qual_name = (
                table_process.name +
                "@derived_column:{}".format(
                    row[target_header["column"]]
                ))

            _proc_inputs = [] if source_qual_name not in columnEntitiesOutput else [
                columnEntitiesOutput[source_qual_name].to_json(minimum=True)]
            _proc_output = [
                columnEntitiesOutput[target_qual_name].to_json(minimum=True)]

            process_entity = AtlasProcess(
                name=table_process.name,
                typeName=process_type,
                # qualifiedName can be overwritten via the
                # attributes functionality
                qualified_name=process_qual_name,
                guid=self.guidTracker.get_guid(),
                # Assuming always a single output
                inputs=_proc_inputs,
                outputs=_proc_output,
                attributes=process_attributes,
                # TODO: Make the relationship name more dynamic instead of
                # hard coding query
                relationshipAttributes={
                    "query": table_process.to_json(minimum=True)}
            )
            if process_entity in columnEntitiesOutput:
                popped_entity = columnEntitiesOutput.pop(process_entity)
                process_entity.merge(popped_entity)
            # Add to outputs
            columnEntitiesOutput.update(
                {process_entity.qualifiedName: process_entity})

            if use_column_mapping:
                # Handles multiple source columns from multiple source datasets
                col_map_source_col = (
                    "*" if source_qual_name is None
                    else columnEntitiesOutput[source_qual_name].name
                )
                col_map_target_col = (
                    columnEntitiesOutput[target_qual_name].name
                )
                col_map_source_table = row[source_header["table"]] or "*"
                col_map_target_table = row[target_header["table"]]
                hash_key = hash(col_map_source_table + col_map_target_table)
                col_map_dict = {"Source": col_map_source_col,
                                "Sink": col_map_target_col}
                data_map_dict = {"Source": col_map_source_table,
                                 "Sink": col_map_target_table}

                if table_process.guid in dataset_mapping:
                    if hash_key in dataset_mapping[table_process.guid]:
                        # Hash Key has not been seen before
                        (
                            dataset_mapping[table_process.guid][hash_key]["ColumnMapping"]
                            .append(col_map_dict)
                        )
                    else:
                        # Hash key has been seen before
                        dataset_mapping[table_process.guid][hash_key] = {
                            "ColumnMapping": [col_map_dict],
                            "DatasetMapping": data_map_dict
                        }
                else:
                    # This guid has never been seen before
                    dataset_mapping[table_process.guid] = {
                        hash_key: {
                            "ColumnMapping": [col_map_dict],
                            "DatasetMapping": data_map_dict
                        }
                    }
        # Update the passed in atlas_entities if we are using column mapping
        if use_column_mapping:
            for entity in atlas_entities:
                if entity.guid in dataset_mapping:
                    # hash_key: {DSMap:{}, ColumnMapping}
                    column_mapping_attribute = [
                        mappings for mappings in dataset_mapping[entity.guid].values()]
                    entity.attributes.update(
                        {"columnMapping": json.dumps(column_mapping_attribute)}
                    )
        output = [e for e in list(columnEntitiesOutput.values())]
        return output

    def _determine_dataset_to_use(self, qual_name, typeName):
        results = []
        if qual_name is None:
            # This will allow us to do a partial update.
            results = None
        elif qual_name == "N/A":
            # N/A is a special keyword that forces us to override
            # and delete the existing input/output
            results = []
        else:
            # This is an AtlasObjectId, necessary if we don't have the
            # guid of the existing object or have it as a referenced object
            results = [{"typeName": typeName,
                        "uniqueAttributes": {"qualifiedName": qual_name}}]
        return results

    def _header_qn(self, l):
        return l["uniqueAttributes"]["qualifiedName"]

    def parse_update_lineage(self, json_rows):
        """
        Take in UpdateLineage dictionaries and create the mutated Process
        entities to be uploaded. All referenced entities must already exist
        and be identified by their type and qualifiedName.

        Assumes a None entry for target or source qualifiedNames means 
        "no change" to the existing entity. Using 'N/A' for the target or
        source qualifiedNames will reset the existing input or output to an
        empty list.

        :param json_rows:
            A list of dicts that contain the converted rows of your update
            lineage spreadsheet.
        :type json_rows: list(dict(str,str))
        :return:
            A list of Atlas Processes as dictionaries representing the updated
            process entity.
        :rtype: list(dict)
        """
        results = []
        sp = self.config.source_prefix
        tp = self.config.target_prefix
        pp = self.config.process_prefix

        processes_seen = dict()

        for row in json_rows:
            try:
                target_type = row[f"{tp} typeName"]
                target_qual_name = row[f"{tp} qualifiedName"]
                source_type = row[f"{sp} typeName"]
                source_qual_name = row[f"{sp} qualifiedName"]
                process_type = row[f"{pp} typeName"]
                process_qual_name = row[f"{pp} qualifiedName"]
                process_name = row[f"{pp} name"]
            except KeyError:
                raise Exception(json.dumps(row))

            # Determine whether this should destroy one side, partial update
            # (one side), or full update (both sides).
            inputs = self._determine_dataset_to_use(
                source_qual_name, source_type)
            outputs = self._determine_dataset_to_use(
                target_qual_name, target_type)

            if process_qual_name in processes_seen:
                temp_proc = processes_seen[process_qual_name]
                # Get the inputs and outputs as they exist before this row
                temp_in = temp_proc.inputs
                temp_out = temp_proc.outputs
                # If we have an input, check if it already exists
                if inputs and len(inputs) > 0:
                    input_qn = self._header_qn(inputs[0])    
                    if input_qn not in [self._header_qn(x) for x in temp_in]:
                        temp_in.extend(inputs)
                        temp_proc.inputs = temp_in
                        
                    else:
                        warnings.warn(f"Input '{input_qn}' is repeated in Process '{process_qual_name}'. Only the earliest entry is kept.")
                elif isinstance(inputs, list):
                    # We have an empty list, as the input, meaning destroy the input
                    if len(temp_in) > 0:
                        warnings.warn(f"Process '{process_qual_name}' has conflicting inputs and N/A values and will possibly be overwritten.")
                    temp_proc.inputs(inputs)

                if outputs:
                    output_qn = self._header_qn(outputs[0])    
                    if output_qn not in [self._header_qn(x) for x in temp_out]:
                        temp_out.extend(outputs)
                        temp_proc.outputs = temp_out
                    else:
                        warnings.warn(f"Output '{output_qn}' is repeated in Process '{process_qual_name}'. Only the earliest entry is kept.")
                elif isinstance(outputs, list):
                    # We have an empty list, as the output, meaning destroy the output
                    if len(temp_out) > 0:
                        warnings.warn(f"Process '{process_qual_name}' has conflicting outputs and N/A values and will possibly be overwritten.")
                    temp_proc.outputs = outputs
                
            else:
                proc = AtlasProcess(
                    name=process_name,
                    typeName=process_type,
                    qualified_name=process_qual_name,
                    guid=self.guidTracker.get_guid(),
                    inputs=inputs,
                    outputs=outputs
                )
                processes_seen[process_qual_name] = proc
            
            results = [v.to_json() for v in processes_seen.values()]
        return results



from warnings import warn
from collections import OrderedDict

from ..core.util import GuidTracker
from ..core import (
    AtlasAttributeDef,
    AtlasEntity,
    EntityTypeDef
)

from .lineagemixin import LineageMixIn
from . import util as reader_util


class ReaderConfiguration():
    """
    A base configuration for the Reader class.  Allows you to customize
    headers with a source_prefix, target_prefix, and process_prefix for
    parsing table and column lineages.
    """

    def __init__(self, **kwargs):
        super().__init__()
        self.value_separator = kwargs.get('value_separator', ';')
        self.source_prefix = kwargs.get(
            "source_prefix", "Source")
        self.target_prefix = kwargs.get(
            "target_prefix", "Target")
        self.process_prefix = kwargs.get(
            "process_prefix", "Process")
        self.column_transformation_name = kwargs.get(
            "column_transformation_name", "transformation")


class Reader(LineageMixIn):
    """
    The base Reader with functionality that supports python dicts.
    """
    TEMPLATE_HEADERS = {
        "ColumnsLineage": [
            "Target table", "Target column", "Target classifications",
            "Source table", "Source column", "Source classifications",
            "transformation"
        ],
        "TablesLineage": [
            "Target table", "Target type", "Target classifications",
            "Source table", "Source type", "Source classifications",
            "Process name", "Process type"
        ],
        "EntityDefs": [
            "Entity TypeName", "name", "description",
            "isOptional", "isUnique", "defaultValue",
            "typeName", "displayName", "valuesMinCount",
            "valuesMaxCount", "cardinality", "includeInNotification",
            "indexType", "isIndexable"
        ],
        "BulkEntities": [
            "typeName", "name", "qualifiedName", "classifications"
        ],
        "UpdateLineage": [
            "Target typeName", "Target qualifiedName", "Source typeName",
            "Source qualifiedName", "Process name", "Process qualifiedName",
            "Process typeName"
        ]
    }

    def __init__(self, configuration, guid=-1000):
        """
        Creates the base Reader with functionality that supports python dicts.

        :param configuration:
            A list of dicts containing at least `Entity TypeName` and `name`
        :type configuration:
            :class:`~pyapacheatlas.readers.reader.ReaderConfiguration`
        :param int guid:
            A negative integer to use as the starting counter for entities
            created by this reader.
        """
        super().__init__()
        self.config = configuration
        self.guidTracker = GuidTracker(guid)

    def _organize_attributes(self, row, existing_entities, ignore=[]):
        """
        Organize the row entries into a distinct set of attributes and
        relationshipAttributes.

        :param dict(str,str) row:
            A dict representing the input rows.
        :param existing_entities:
            A list of existing atlas entities that will be used to infer
            any relationship attributes.
        :type existing_entities:
            dict(str, `:class:~pyapacheatlas.core.entity.AtlasEntity`)
        :param list(str) ignore:
            A set of keys to ignore and omit from the returned dict.
        :return:
            A dictionary containing 'attributes' and 'relationshipAttributes'
        :rtype: dict(str, dict(str,str))
        """
        output = {"attributes": {}, "relationshipAttributes": {}}
        for k, v in row.items():
            # Remove the required attributes so they're not double dipping.
            if k in ignore:
                continue
            # Remove any cell with a None / Null attribute
            elif v is None:
                continue
            # If the Attribute key starts with [Relationship]
            # Move it to the relation
            elif k.startswith("[Relationship]"):
                cleaned_key = k.replace("[Relationship]", "").strip()
                # Assuming that we can find this in an existing entity
                try:
                    min_reference = existing_entities[v].to_json(minimum=True)
                # LIMITATION: We must have already seen the relationship
                # attribute to be certain it can be looked up.
                except KeyError:
                    raise KeyError(
                        f"The entity {v} should be listed before {row['qualifiedName']}."
                    )
                output["relationshipAttributes"].update(
                    {cleaned_key: min_reference}
                )
            else:
                output["attributes"].update({k: v})

        return output

    def parse_bulk_entities(self, json_rows):
        """
        Create an AtlasTypeDef consisting of entities and their attributes
        for the given json_rows.

        :param list(dict(str,str)) json_rows:
            A list of dicts containing at least `Entity TypeName` and `name`
            that represents the metadata for a given entity type's attributeDefs.
            Extra metadata will be ignored.
        :return: An AtlasTypeDef with entityDefs for the provided rows.
        :rtype: dict(str, list(dict))
        """
        # For each row,
        # Extract the
        # Extract any additional attributes
        req_attribs = ["typeName", "name", "qualifiedName", "classifications", "owners", "experts"]
        existing_entities = OrderedDict()

        for row in json_rows:

            if ((row["name"] is None) or (row["typeName"] is None) or
                    (row["qualifiedName"] is None)):
                # An empty row snuck in somehow, skip it.
                continue

            _attributes = self._organize_attributes(
                row,
                existing_entities,
                req_attribs
            )

            entity = AtlasEntity(
                name=row["name"],
                typeName=row["typeName"],
                qualified_name=row["qualifiedName"],
                guid=self.guidTracker.get_guid(),
                attributes=_attributes["attributes"],
                classifications=reader_util.string_to_classification(
                    row["classifications"],
                    sep=self.config.value_separator
                ),
                relationshipAttributes=_attributes["relationshipAttributes"]
            )
            if "experts" in row or "owners" in row and len( row.get("experts", []) + row.get("owners", []) ) > 0:
                experts = []
                owners = []
                if len(row.get("experts", []) or [])>0:
                    experts = [{"id":e} for e in row.get("experts", "").split(self.config.value_separator) if e != '']
                if len(row.get("owners", []) or [])>0:
                    owners = [{"id":o} for o in row.get("owners", "").split(self.config.value_separator) if o != '']
                entity.contacts = {"Expert": experts, "Owner": owners }

            existing_entities.update({row["qualifiedName"]: entity})

        output = {"entities": [e.to_json()
                               for e in list(existing_entities.values())]}
        return output

    def parse_entity_defs(self, json_rows):
        """
        Create an AtlasTypeDef consisting of entityDefs for the
        given json_rows.

        :param list(dict(str,str)) json_rows:
            A list of dicts containing at least `Entity TypeName` and `name`
            that represents the metadata for a given entity type's
            attributeDefs. Extra metadata will be ignored.
        :return: An AtlasTypeDef with entityDefs for the provided rows.
        :rtype: dict(str, list(dict))
        """
        entities = dict()
        attribute_metadata_seen = set()
        output = {"entityDefs": []}
        # Required attributes
        # Get all the attributes it's expecting official camel casing
        # with the exception of "Entity TypeName"
        for row in json_rows:
            try:
                entityTypeName = row["Entity TypeName"]
            except KeyError:
                raise KeyError("Entity TypeName not found in {}".format(row))

            _ = row.pop("Entity TypeName")
            # Update all seen attribute metadata
            columns_in_row = list(row.keys())
            attribute_metadata_seen = attribute_metadata_seen.union(
                set(columns_in_row))
            # Remove any null cells, otherwise the AttributeDefs constructor
            # doesn't use the defaults.
            for column in columns_in_row:
                if row[column] is None:
                    _ = row.pop(column)

            json_entity_def = AtlasAttributeDef(**row).to_json()

            if entityTypeName not in entities:
                entities[entityTypeName] = []

            entities[entityTypeName].append(json_entity_def)

        # Create the entitydefs
        for entityType in entities:
            local_entity_def = EntityTypeDef(
                name=entityType,
                attributeDefs=entities[entityType],
                # Adding this as a default until I figure
                # do this from the excel / json readers.
                superTypes=["DataSet"]
            ).to_json()
            output["entityDefs"].append(local_entity_def)

        # Extra attribute metadata (e.g. extra columns / json entries)
        # are ignored. Warn the user that this metadata will be ignored.
        extra_metadata_warnings = [
            i for i in attribute_metadata_seen if i not in AtlasAttributeDef.propertiesEnum]
        for extra_metadata in extra_metadata_warnings:
            warn(("The attribute metadata \"{}\" is not a part of the Atlas" +
                  " Attribute Def and will be ignored.").format(
                extra_metadata))

        return output

    @staticmethod
    def make_template():
        """
        Generate a template for the given reader.
        """
        raise NotImplementedError



from datetime import datetime
import json
import requests

from .base import AtlasAuthBase


class ServicePrincipalAuthentication(AtlasAuthBase):
    """
    Authenticates to the Azure OAuth provider using a service principal.
    """

    def __init__(self, tenant_id, client_id, client_secret):
        """
        :param str tenant_id: The tenant id of your Azure subscription.
        :param str client_id: The client id or application id of your
            service principal.
        :param str client_secret: The client secret or application secret
            of your service principal.
        """
        super().__init__()

        self.ouath_url = (
            "https://login.microsoftonline.com/" +
            tenant_id + "/oauth2/token"
        )
        self.data = {
            # This is the resource id for the data catalog
            "resource": "73c2949e-da2d-457a-9607-fcc665198967",
            "client_id": client_id,
            "grant_type": "client_credentials",
            "client_secret": client_secret
        }
        self.access_token = None
        self.expiration = datetime.now()

    def _set_access_token(self):
        """
        Sets the access token for your session.
        """
        authResponse = requests.post(self.ouath_url, data=self.data)
        if authResponse.status_code != 200:
            authResponse.raise_for_status()

        authJson = json.loads(authResponse.text)

        self.access_token = authJson["access_token"]
        self.expiration = datetime.fromtimestamp(int(authJson["expires_on"]))

    def get_authentication_headers(self):
        """
        Gets the current access token or refreshes the token if it
        has expired.
        :return: The authorization headers.
        :rtype: dict(str, str)
        """
        if self.expiration <= datetime.now():
            self._set_access_token()

        return {
            "Authorization": "Bearer " + self.access_token,
            "Content-Type": "application/json"
        }



from enum import Enum


class TypeCategory(Enum):
    """
    An implementation of an Atlas TypeCategory used in relationshipDefs.
    """
    CLASSIFICATION = "classification"
    ENTITY = "entity"
    ENUM = "enum"
    RELATIONSHIP = "relationship"
    STRUCT = "struct"


class Cardinality(Enum):
    """
    An implementation of an Atlas Cardinality used in relationshipDefs.
    """
    SINGLE = "SINGLE"
    LIST = "LIST"
    SET = "SET"


class AtlasAttributeDef():
    """
    An implementation of AtlasAttributeDef.

    :param str name:
        The name of the Attribute Definition. Provides a standard way to pass
        in an attribute definition when defining an Entity.

    Kwargs:
        :param cardinality:
            One of Cardinality.SINGLE, .SET, .LIST. Defaults to SINGLE.
        :type cardinality: :class:`pyapacheatlas.core.typedef.Cardinality`
        :param str typeName: The type of this attribute. Defaults to string.
    """

    propertiesEnum = [
        "cardinality", "constraints", "defaultValue", "description",
        "displayName", "includeInNotification", "indexType", "isIndexable",
        "isOptional", "isUnique", "name", "options", "searchWeight",
        "typeName", "valuesMaxCount", "valuesMinCount"
    ]

    def __init__(self, name, **kwargs):
        """
        Default arguments are chosen assuming you want a single attribute
        """
        super().__init__()
        self.cardinality = kwargs.get(
            "cardinality", Cardinality.SINGLE.value)  # Cardinality
        # array of AtlasConstraintDef
        self.constraints = kwargs.get("constraints")
        self.defaultValue = kwargs.get("defaultValue")  # string
        self.description = kwargs.get("description")  # string
        self.displayName = kwargs.get("displayName")  # string
        self.includeInNotification = kwargs.get(
            "includeInNotification", False)  # boolean
        self.indexType = kwargs.get("indexType")  # IndexType
        self.isIndexable = kwargs.get("isIndexable", False)  # boolean
        self.isOptional = kwargs.get("isOptional", True)  # boolean
        self.isUnique = kwargs.get("isUnique", False)  # boolean
        self.name = name  # string
        self.options = kwargs.get("options")  # map of string
        self.searchWeight = kwargs.get("searchWeight")  # number
        self.typeName = kwargs.get("typeName", "string")  # string
        self.valuesMaxCount = kwargs.get("valuesMaxCount", 1)  # number
        self.valuesMinCount = kwargs.get("valuesMinCount", 0)  # number

    def to_json(self, omit_nulls=True):
        output = self.__dict__
        if omit_nulls:
            output = {k: v for k, v in output.items(
            ) if v is not None and omit_nulls}
        return output

class AtlasRelationshipAttributeDef(AtlasAttributeDef):
    """
    An implementation of AtlasRelationshipAttributeDef. Provides a standard
    way to pass in a relationship definition when defining an Entity rather
    than creating a separate relationship def.

    :param str name: The name of the Relationship Attribute Definition.
    :param str relationshipTypeName:
        The name of the relationship type being defined. Commonly uses
        'endDef1_endDef2' where endDef's are the names given based on
        the relationship being used.

    Kwargs:
        :param cardinality:
            One of Cardinality.SINGLE, .SET, .LIST. Defaults to SINGLE.
        :type cardinality: :class:`pyapacheatlas.core.typedef.Cardinality`
        :param str typeName: The type of this attribute. Defaults to string.
    """
    def __init__(self, name, relationshipTypeName, **kwargs):
        super().__init__(name, **kwargs)
        self.relationshipTypeName = relationshipTypeName


class BaseTypeDef():
    """
    An implementation of AtlasBaseTypeDef.
    
    :param str name: The name of the typedef.
    :param category: The category of the typedef.
    :type category: :class:`~pyapacheatlas.core.typedef.TypeCategory`
    """

    def __init__(self, name, category, **kwargs):
        
        super().__init__()
        self.category = category.value.upper()
        self.createTime = kwargs.get("createTime")
        self.createdBy = kwargs.get("createdBy")
        self.dateFormatter = kwargs.get("dateFormatter")
        self.description = kwargs.get("description")
        self.guid = kwargs.get("guid")
        self.name = name
        self.options = kwargs.get("options")
        self.serviceType = kwargs.get("serviceType")
        self.typeVersion = kwargs.get("typeVersion")
        self.updateTime = kwargs.get("updateTime")
        self.updatedBy = kwargs.get("updatedBy")
        self.version = kwargs.get("version")

    def to_json(self, omit_nulls=True):
        """
        Converts the typedef object to a dict / json.

        :param bool omit_null: If True, omits keys with value of None.
        :return: The dict / json version of the type def.
        :rtype: dict
        """
        output = self.__dict__
        if omit_nulls:
            output = {k: v for k, v in output.items(
            ) if v is not None and omit_nulls}
        return output


class AtlasStructDef(BaseTypeDef):
    """
    An implemention of AtlasStructDef. Not expected to be used by the end users.

    :param str name: The name of the type definition.
    :param category: The category of the typedef.
    :type category: :class:`~pyapacheatlas.core.typedef.TypeCategory`

    Kwargs:
        :param attributeDefs:
            The AtlasAttributeDefs that should be available on the struct.
        :type attributeDefs: list(Union(dict, :class:`pyapacheatlas.core.typedef.AtlasAttributeDef`))
    """

    def __init__(self, name, category, **kwargs):
        super().__init__(name=name, category=category, **kwargs)
        self.attributeDefs = kwargs.get("attributeDefs", [])

    @property
    def attributeDefs(self):
        """
        :return: List of attribute definitions.
        :rtype: list(dict)
        """
        return self._attributeDefs

    @attributeDefs.setter
    def attributeDefs(self, value):
        """
        :param value:
            The attribute defs you are adding. They are comma delimited dicts
            or AtlasAttributeDefs.
        :type value: list(Union(dict, :class:`pyapacheatlas.core.typedef.AtlasAttributeDef`))
        """
        self._attributeDefs = [
            e.to_json()
            if isinstance(e, AtlasAttributeDef)
            else e
            for e in value
        ]

    def addAttributeDef(self, *args):
        """
        Add one or many attribute definitions.

        :param args:
            The attribute defs you are adding. They are comma delimited dicts
            or AtlasAttributeDefs. You can expand a list with `*my_list`.
        :type args: Union(dict, :class:`pyapacheatlas.core.typedef.AtlasAttributeDef`)
        """
        self.attributeDefs = self.attributeDefs + [
            e.to_json()
            if isinstance(e, AtlasAttributeDef)
            else e
            for e in args
        ]

    def to_json(self, omit_nulls=True):
        output = super().to_json(omit_nulls)
        output.update({"attributeDefs": self.attributeDefs})
        output.pop("_attributeDefs")
        return output


class ClassificationTypeDef(AtlasStructDef):
    """
    An implementation of AtlasClassificationDef

    :param str name: The name of the type definition.
    :param list(str) entityTypes: The list of entityTypes for the classification.
    :param list(str) superTypes: The list of superTypes for the classification.

    Kwargs:
        :param attributeDefs:
            The AtlasAttributeDefs that should be available on the Classification.
        :type attributeDefs: list(Union(dict, :class:`pyapacheatlas.core.typedef.AtlasAttributeDef`))
        :param list(str) subTypes: The types that will inherit this classification.
    """

    def __init__(self, name, entityTypes=[], superTypes=[], **kwargs):
        super().__init__(name, category=TypeCategory.CLASSIFICATION, **kwargs)
        self.entityTypes = entityTypes
        self.superTypes = superTypes
        self.subTypes = kwargs.get("subTypes", []) or []

    def __str__(self):
        return self.name


class EntityTypeDef(AtlasStructDef):
    """
    An implementation of AtlasEntityDef

    :param str name: The name of the type definition.
    :param list(str) superTypes:
        The list of superTypes for the classification. You most likely want
        ['DataSet'] to create a DataSet asset which is the default.

    Kwargs:
        :param attributeDefs:
            The AtlasAttributeDefs that should be available on the Entity.
        :type attributeDefs: list(Union(dict, :class:`pyapacheatlas.core.typedef.AtlasAttributeDef`))
    """

    def __init__(self, name, superTypes=['DataSet'], **kwargs):
        super().__init__(name, category=TypeCategory.ENTITY, **kwargs)
        self.relationshipAttributeDefs = kwargs.get(
            "relationshipAttributeDefs", []) or []
        self.superTypes = superTypes

    def __str__(self):
        return self.name
    
    @property
    def relationshipAttributeDefs(self):
        """
        :return: List of relationship attribute definitions.
        :rtype: list(dict)
        """
        return self._relationshipAttributeDefs

    @relationshipAttributeDefs.setter
    def relationshipAttributeDefs(self, value):
        """
        :param value:
            The attribute defs you are adding. They are comma delimited dicts
            or AtlasRelationshipAttributeDefs.
        :type value: list(Union(dict, :class:`pyapacheatlas.core.typedef.AtlasRelationshipAttributeDef`))
        """
        self._relationshipAttributeDefs = [
            e.to_json()
            if isinstance(e, AtlasRelationshipAttributeDef)
            else e
            for e in value
        ]

    def addRelationshipAttributeDef(self, *args):
        """
        Add one or many attribute definitions.

        :param args:
            The attribute defs you are adding. They are comma delimited dicts
            or AtlasAttributeDefs. You can expand a list with `*my_list`.
        :type args: Union(dict, :class:`pyapacheatlas.core.typedef.relationshipAttributeDefs`)
        """
        self.relationshipAttributeDefs = self.relationshipAttributeDefs + [
            e.to_json()
            if isinstance(e, AtlasRelationshipAttributeDef)
            else e
            for e in args
        ]
    
    def to_json(self, omit_nulls=True):
        output = super().to_json(omit_nulls)
        output.update({"relationshipAttributeDefs": self.relationshipAttributeDefs})
        output.pop("_relationshipAttributeDefs")
        return output


class RelationshipTypeDef(BaseTypeDef):
    """
    An implementation of AtlasRelationshipDef.

    :param str name: The name of the relationship type def.
    :param endDef1:
        Either a valid AtlasRelationshipEndDef dict or class object.
    :type endDef1:
        Union(:class:`~pyapacheatlas.core.typedef.AtlasRelationshipEndDef`, dict)
    :param endDef2:
        Either a valid AtlasRelationshipEndDef dict or class object.
    :type endDef2:
        Union(:class:`~pyapacheatlas.core.typedef.AtlasRelationshipEndDef`, dict)
    :param str relationshipCategory:
        One of COMPOSITION, AGGREGATION, ASSOCIATION. You're most likely
        looking at COMPOSITION to create a parent/child relationship.
    """

    def __init__(self, name, endDef1, endDef2, relationshipCategory, **kwargs):
        super().__init__(name, category=TypeCategory.RELATIONSHIP, **kwargs)

        self.endDef1 = endDef1
        self.endDef2 = endDef2
        self.relationshipCategory = relationshipCategory

    @property
    def endDef1(self):
        """
        :return: The first end definition.
        :rtype: dict
        """
        return self._endDef1

    @endDef1.setter
    def endDef1(self, value):
        """
        :param value:
            The first end def for you are adding. They are comma delimited dicts
            or AtlasRelationshipEndDef.
        :type value: list(Union(dict, :class:`pyapacheatlas.core.typedef.AtlasRelationshipEndDef`))
        """
        if isinstance(value, AtlasRelationshipEndDef):
            self._endDef1 = value.to_json()
        elif isinstance(value, dict):
            self._endDef1 = value
        else:
            raise NotImplementedError(
                f"An EndDef of type `{type(value)}` is not supported.")

    @property
    def endDef2(self):
        """
        :return: The second end definition.
        :rtype: dict
        """
        return self._endDef2

    @endDef2.setter
    def endDef2(self, value):
        """
        :param value:
            The second end def for you are adding. They are comma delimited
            dicts or AtlasRelationshipEndDef.
        :type value: list(Union(dict, :class:`pyapacheatlas.core.typedef.AtlasRelationshipEndDef`))
        """
        if isinstance(value, AtlasRelationshipEndDef):
            self._endDef2 = value.to_json()
        elif isinstance(value, dict):
            self._endDef2 = value
        else:
            raise NotImplementedError(
                f"An EndDef of type `{type(value)}` is not supported.")

    def to_json(self, omit_nulls=True):
        output = super().to_json(omit_nulls)
        output.update({"endDef1": self.endDef1, "endDef2": self.endDef2})
        output.pop("_endDef1")
        output.pop("_endDef2")
        return output


class AtlasRelationshipEndDef():
    """
    An implementation of AtlasRelationshipEndDef.

    :param str name: The name that will appear on the entity's relationship attribute.
    :param str typeName: The type that is required for this end of the relationship.
    :param cardinality: The cardinality of the end definition.
    :type cardinality": :class:`~pyapacheatlas.core.typedef.Cardinality`
    :param bool isContainer:
        This should be False when the cardinality is SINGLE. It should be
        True when cardinality is SET or LIST. endDef1 should

    Kwargs:
        :param str description:
            The description of this end of the relationship.
        :param bool isLegacyAttribute: Defaults to False.
    """

    def __init__(self, name, typeName, cardinality=Cardinality.SINGLE, isContainer=False, **kwargs):
        self.cardinality = cardinality.value
        self.description = kwargs.get("description")
        self.isLegacyAttribute = kwargs.get("isLegacyAttribute", False)
        self.name = name
        self.type = typeName
        self.isContainer = isContainer

    def to_json(self, omit_nulls=True):
        """
        Converts the typedef object to a dict / json.

        :param bool omit_null: If True, omits keys with value of None.
        :return: The dict / json version of the type def.
        :rtype: dict
        """
        output = self.__dict__
        if omit_nulls:
            output = {k: v for k, v in output.items(
            ) if v is not None and omit_nulls}
        return output


class ParentEndDef(AtlasRelationshipEndDef):
    """
    A helper for creating a Parent end def (e.g. EndDef1 that is a container).
    The goal being to simplify and reduce the margin of error when creating
    containing relationships. This should be used in EndDef1 when the 
    relationshipCategory is COMPOSITION or AGGREGATION.

    Defaults to cardinality of SET and isContainer=True.
    """

    def __init__(self, name, typeName, **kwargs):
        super().__init__(
            name, typeName,
            cardinality=Cardinality.SET,
            isContainer=True, **kwargs
        )


class ChildEndDef(AtlasRelationshipEndDef):
    """
    A helper for creating a Child end def (e.g. EndDef2 that is a single).
    The goal being to simplify and reduce the margin of error when creating
    containing relationships. This should be used in EndDef2 when the 
    relationshipCategory is COMPOSITION or AGGREGATION.

    Defaults to cardinality of SINGLE and isContainer=False.
    """

    def __init__(self, name, typeName, **kwargs):
        super().__init__(
            name, typeName,
            cardinality=Cardinality.SINGLE,
            isContainer=False, **kwargs
        )



import warnings


class AtlasEntity():
    """
    A python representation of the AtlasEntity from Apache Atlas.

    :param str name: The name of this instance of an atlas entity.
    :param str typeName: The type this entity should be.
    :param str qualified_name: The unique "qualified name" of this
        instance of an atlas entity.
    :param Union(str,int) guid:
        The guid to reference this entity by. Should be a negative number
        if you're adding an entity. Consider using get_guid() method from
        :class:`~pyapacheatlas.core.util.GuidTracker` to retrieve unique
        negative numbers.
    :param dict, optional relationshipAttributes: The relationship attributes
        representing how this entity is connected to others.  Commonly
        used for "columns" to indicate entity is a column of a table or
        "query" to indicate a process entity is tied another process in
        a column lineage scenario.
    :param dict, optional attributes: Additional attributes that your
        atlas entity may require.
    :param dict, optional classifications: Classifications that may
        be applied to this atlas entity.
    :param dict(str, dict(str, list(dict(strt,str)))), optional contacts:
        Contacts should contain keys Experts and/or Owners. Their values should
        be a list of dicts with keys id and info. Id is a microsoft graph
        object id. Info is a string of extra information.
    """

    def __init__(self, name, typeName, qualified_name, guid=None, **kwargs):
        super().__init__()
        self.typeName = typeName
        self.guid = guid
        self.attributes = kwargs.get("attributes", {})
        self.attributes.update({"name": None, "qualifiedName": None})
        self.name = name
        self.qualifiedName = qualified_name
        if "description" in kwargs:
            self.attributes.update({"description": kwargs["description"]})
        self.relationshipAttributes = kwargs.get("relationshipAttributes", {})
        self.classifications = kwargs.get("classifications", None)
        # This isn't implemented in Apache Atlas, so being cautious
        if "contacts" in kwargs:
            # Data Structure: {"Expert":[{"id","info"}], "Owner":...}
            self.contacts = kwargs.get("contacts", {})

    def __eq__(self, other):
        return self.qualifiedName == other

    def __hash__(self):
        return hash(self.qualifiedName)

    def __ne__(self, other):
        return self.qualifiedName != other

    def __repr__(self):
        return "AtlasEntity({type_name},{qual_name})".format(
            type_name=self.typeName,
            qual_name=self.qualifiedName
        )

    def __str__(self):
        return "AtlasEntity({type_name},{qual_name})".format(
            type_name=self.typeName,
            qual_name=self.qualifiedName
        )

    @property
    def name(self):
        """
        Retrieve the name of this entity.

        :return: The name of the entity.
        :rtype: str
        """
        return self.attributes["name"]

    @name.setter
    def name(self, value):
        """
        Set the name of this entity.

        :param str value: The name of the entity.
        """
        self.attributes["name"] = value

    @property
    def qualifiedName(self):
        """
        Retrieve the qualifiedName of this entity.

        :return: The name of the entity.
        :rtype: str
        """
        return self.attributes["qualifiedName"]

    @qualifiedName.setter
    def qualifiedName(self, value):
        """
        Set the qualifiedName of this entity.

        :param str value: The qualifiedName of the entity.
        """
        self.attributes["qualifiedName"] = value

    def addRelationship(self, **kwargs):
        """
        Add one or many relationshipAttributes to the outputs. This will
        also update an existing attribute. You can pass in a parameter name and
        then either an Atlas Entity or a dict. The dict is a 
        table = AtlasEntity(...) or

        Kwargs:
            :param kwarg: The name of the relationship attribute you're adding.
            :type kwarg:
                Union(dict, :class:`pyapacheatlas.core.entity.AtlasEntity`)
        """
        for k,v in kwargs.items():
            val = v.to_json(minimum=True) if isinstance(v, AtlasEntity) else v
            self.relationshipAttributes.update({k: val })

    def get_name(self):
        """
        Deprecated in favor of .name property.
        Retrieve the name of this entity.

        :return: The name of the entity.
        :rtype: str
        """
        warnings.warn("Get name using AtlasEntity.name.",
                      category=DeprecationWarning, stacklevel=2)
        return self.attributes["name"]

    def get_qualified_name(self):
        """
        Deprecated in favor of .qualifiedName.
        Retrieve the qualified (unique) name of this entity.

        :return: The qualified name of the entity.
        :rtype: str
        """
        warnings.warn("Get qualified name using AtlasEntity.qualifiedName.",
                      category=DeprecationWarning, stacklevel=2)
        return self.attributes["qualifiedName"]

    def to_json(self, minimum=False):
        """
        Convert this atlas entity to a dict / json. Returns typename, guid,
        and qualified name if guid is not none. If guid is None then this will
        return typename, uniqueAttributes with a sub object of qualified name.

        By specifying a guid, this method assumes you will be uploading the
        entity (and want or at least willing to accept changes to the entity).
        By NOT specifying a guid, this assumes you will be using the entity as
        a reference used by another one in the upload (e.g. creating a process
        entity that uses an existing entity as an input or output).

        :param bool minimum: If True, returns only the
            type name, qualified name, and guid of the entity (when guid is
            defined). If True and guid is None, returns typeName,
            uniqueAttributes and qualifiedName. If False, return the full entity
            and its attributes and relationship attributes.
        :return: The json representation of this atlas entity.
        :rtype: dict
        """
        if minimum and self.guid is not None:
            output = {
                "typeName": self.typeName,
                "guid": self.guid,
                "qualifiedName": self.attributes["qualifiedName"]
            }
        elif minimum and self.guid is None:
            output = {
                "typeName": self.typeName,
                "uniqueAttributes": {
                    "qualifiedName": self.qualifiedName
                }
            }
        else:
            output = {
                "typeName": self.typeName,
                "guid": self.guid,
                "attributes": self.attributes,
                "relationshipAttributes": self.relationshipAttributes
            }
            # Add ins for optional top level attributes
            if self.classifications:
                output.update({"classifications": self.classifications})
            if hasattr(self, 'contacts'):
                output.update({"contacts": self.contacts})


        return output

    def merge(self, other):
        """
        Update the calling object with the attributes and classifications of
        the passed in AtlasEntity.

        :param :class:`~pyapacheatlas.core.entity.AtlasEntity` other:
            The other AtlasEntity object that you want to merge.
        """
        if self.qualifiedName != other.qualifiedName:
            raise TypeError("Type:{} cannot be merged with {}".format(
                type(other), type(self)))

        # Take the "earlier" defined entity's guid
        self.guid = other.guid
        # Add attributes that are not present in the later row's attributes
        # Meaning, later attributes SUPERCEDE attributes
        # This helps with updating input and output attributes
        # later in process entities.
        _other_attr_keys = set(other.attributes.keys())
        _self_attr_keys = set(self.attributes.keys())
        _new_keys_in_other = _other_attr_keys.difference(_self_attr_keys)
        self.attributes.update(
            {k: v for k, v in other.attributes.items()
                if k in _new_keys_in_other})
        # TODO: Handle duplicate classifications
        if other.classifications:
            self.classifications = (self.classifications or []).extend(self.classifications)


class AtlasProcess(AtlasEntity):
    """
    A subclass of AtlasEntity that forces you to include the inputs and
    outputs of the process.

    :param str name: The name of this instance of an atlas entity.
    :param str typeName: The type this entity should be.
    :param str qualified_name: The unique "qualified name" of this
        instance of an atlas entity.
    :param inputs:
        The list of input entities expressed as dicts and in minimum
        format (guid, type name, qualified name) or an AtlasEntity.
    :type inputs: Union(list(dict), :class:`pyapacheatlas.core.entity.EntityDef`)
    :param outputs:
        The list of output entities expressed as dicts and in minimum format
        (guid, type name, qualified name) or an AtlasEntity.
    :type outputs: Union(list(dict), :class:`pyapacheatlas.core.entity.EntityDef`)
    :param Union(str,int), optional guid: The guid to reference this entity by.
    :param dict, optional relationshipAttributes: The relationship attributes
        representing how this entity is connected to others.  Commonly
        used for "columns" to indicate entity is a column of a table or
        "query" to indicate a process entity is tied another process in
        a column lineage scenario.
    :param dict, optional attributes: Additional attributes that your
        atlas entity may require.
    :param dict, optional classifications: Classifications that may
        be applied to this atlas entity.
    """

    def __init__(self, name, typeName, qualified_name, inputs, outputs, guid=None, **kwargs):
        super().__init__(name, typeName, qualified_name, guid=guid, **kwargs)
        self.attributes.update({"inputs": None, "outputs": None})
        self.inputs = inputs
        self.outputs = outputs

    def _parse_atlas_entity(self, iterable):
        """
        :param iterable: An iterable of dict or AtlasEntity
        """
        return [
            e.to_json(minimum=True)
            if isinstance(e, AtlasEntity)
            else e
            for e in iterable
        ]

    @property
    def inputs(self):
        """
        Retrieves the inputs attribute for the process.

        :return: The list of inputs as dicts.
        :rtype: Union(list(dict),None)
        """
        return self.attributes.get("inputs")

    @inputs.setter
    def inputs(self, value):
        """
        Set the inputs attribute for the process. If you pass in a dict list, it
        should be have keys: guid, typeName, qualifiedName. Passing in a list of
        AtlasEntity, it will automatically convert the entities to dicts.

        :param value: List of dicts or atlas entities to set as the inputs.
        :type value: list(Union(dict, :class:`~pyapacheatlas.core.entity.AtlasEntity`))
        """
        # TODO: Consider checking if there is a valid guid to return a simpler min
        if value is not None:
            self.attributes["inputs"] = self._parse_atlas_entity(value)
        else:
            self.attributes["inputs"] = None

    @property
    def outputs(self):
        """
        Retrieves the outputs attribute for the process.

        :return: The list of outputs as dicts.
        :rtype: Union(list(dict),None)
        """
        return self.attributes.get("outputs")

    @outputs.setter
    def outputs(self, value):
        """
        Set the outputs attribute for the process. If you pass in a dict list, it
        should be have keys: guid, typeName, qualifiedName. Passing in a list of
        AtlasEntity, it will automatically convert the entities to dicts.

        :param value: List of dicts or atlas entities to set as the outputs.
        :type value: list(Union(dict, :class:`~pyapacheatlas.core.entity.AtlasEntity`))
        """
        # TODO: Consider checking if there is a valid guid to return a simpler min
        if value is not None:
            self.attributes["outputs"] = self._parse_atlas_entity(value)
        else:
            self.attributes["outputs"] = None

    def addInput(self, *args):
        """
        Add one or many entities to the inputs.

        :param args:
            The atlas entities you are adding. They are comma delimited dicts
            or AtlasEntity. You can expand a list with `*my_list`.
        :type args: Union(dict, :class:`pyapacheatlas.core.entity.AtlasEntity`)
        """
        self.inputs = self.inputs + self._parse_atlas_entity(args)

    def addOutput(self, *args):
        """
        Add one or many entities to the outputs.

        :param args:
            The atlas entities you are adding. They are comma delimited dicts
            or AtlasEntity. You can expand a list with `*my_list`.
        :type args: Union(dict, :class:`pyapacheatlas.core.entity.AtlasEntity`)
        """
        self.outputs = self.outputs + self._parse_atlas_entity(args)

    def merge(self, other):
        """
        Combine the inputs and outputs of a process. Fails if one side has a
        null input or output. Updates the object that merge is called on.

        :param :class:`~pyapacheatlas.core.entity.AtlasEntity` other:
            The other AtlasEntity object that you want to merge.
        """
        super().merge(other)
        # Requires that the input and output attributes have
        # not been altered on self.
        _combined_inputs = self.inputs + other.inputs
        _combined_outputs = self.outputs + other.outputs

        _deduped_inputs = [dict(t) for t in set(
            tuple(d.items()) for d in _combined_inputs)]
        _deduped_outputs = [dict(t) for t in set(
            tuple(d.items()) for d in _combined_outputs)]
        self.inputs = _deduped_inputs
        self.outputs = _deduped_outputs

    def set_inputs(self, inputs):
        """
        Deprecated: Use AtlasEntity.inputs to set inputs.
        Set the inputs to the process.  Inputs should be AtlasEntity minimum
        json of `{qualifiedName:..., guid:..., typeName:...}`.

        :param list(dict) inputs: The minimum json inputs.
        """
        warnings.warn("Set inputs using AtlasProcess.inputs = ...",
                      category=DeprecationWarning, stacklevel=2)
        self.attributes["inputs"] = inputs

    def set_outputs(self, outputs):
        """
        Deprecated: Use AtlasEntity.outputs to set outputs.
        Set the outputs to the process.  Outputs should be AtlasEntity minimum
        json of `{qualifiedName:..., guid:..., typeName:...}`.

        :param list(dict) outputs: The minimum json outputs.
        """
        warnings.warn("Set outputs using AtlasProcess.outputs = ...",
                      category=DeprecationWarning, stacklevel=2)
        self.attributes["outputs"] = outputs

    def get_inputs(self):
        """
        Deprecated: Use AtlasEntity.inputs to get inputs.
        Return the inputs to the process.

        :return: The minimum json inputs.
        :rtype: list(dict)
        """
        warnings.warn("Get inputs using AtlasProcess.inputs.",
                      category=DeprecationWarning, stacklevel=2)
        return self.attributes.get("inputs")

    def get_outputs(self):
        """
        Deprecated: Use AtlasEntity.outputs to set outputs.
        Set the outputs to the process.

        :return: The minimum json inputs.
        :rtype: list(dict)
        """
        warnings.warn("Get outputs using AtlasProcess.outputs.",
                      category=DeprecationWarning, stacklevel=2)
        return self.attributes.get("outputs")


class AtlasClassification():
    """
    A python implementation of the AtlasClassification from Apache Atlas.

    :param str typeName: The name of this classification.
    :param str entityStatus: One of ACTIVE, DELETED, PURGED.
    :param bool propagate:
        Whether the classification should propagate to child entities. Not
        implemented in Purview as of release time.
    :param bool removePropagationsOnEntityDelete:
        Whether the classification should be removed on child entities if the
        parent entity is deleted. Not implemented in Purview as of release time.
    :param dict, optional attributes: Additional attributes that your
        atlas entity may require.
    :param dict, optional validityPeriods: Validity Periods that may
        be applied to this atlas classification.
    """

    def __init__(self, typeName, entityStatus="ACTIVE", propagate=False,
                 removePropagationsOnEntityDelete=False, **kwargs):
        super().__init__()
        if entityStatus not in ["ACTIVE", "PURGED", "DELETED"]:
            raise ValueError(
                f"entityStatus must be one of ACTIVE, PURGED, or DELETED.")
        self.typeName = typeName
        self.entityStatus = entityStatus
        self.propagate = propagate
        self.removePropagationsOnEntityDelete = removePropagationsOnEntityDelete
        self.attributes = kwargs.get("attributes", {})
        self.validityPeriods = kwargs.get("validityPeriods", [])

    def __repr__(self):
        return "AtlasClassification({type_name})".format(
            type_name=self.typeName
        )

    def __str__(self):
        return "AtlasClassification({type_name})".format(
            type_name=self.typeName
        )

    def to_json(self):
        """
        Convert this atlas entity to a dict / json.

        :param bool minimum: If True, returns only the
            type name, qualified name, and guid of the entity.  Useful
            for being referenced in other entities like process inputs
            and outputs.
        :return: The json representation of this atlas entity.
        :rtype: dict
        """

        output = {
            "typeName": self.typeName,
            "entityStatus": self.entityStatus,
            "propagate": self.propagate,
            "removePropagationsOnEntityDelete": self.removePropagationsOnEntityDelete,
            "validityPeriods": self.validityPeriods,
            "attributes": self.attributes
        }
        return output



import setuptools
from pyapacheatlas import __version__

LONG_DESCRIPTION = """
# PyApacheAtlas: API Support for Azure Purview and Apache Atlas

A python package to work with the Apache Atlas API and support bulk loading, custom lineage, and more from a Pythonic set of classes and Excel templates. 

The package currently supports:
* Bulk upload of entities.
* Bulk upload of type definitions.
* Creating custom lineage between two existing entities.
* Creating custom table and complex column level lineage in the [Hive Bridge style](https://atlas.apache.org/0.8.3/Bridge-Hive.html).
  * Supports Azure Purview ColumnMapping Attributes.
* Creating a column lineage scaffolding as in the Hive Bridge Style .
* Performing "What-If" analysis to check if...
   * Your entities are valid types.
   * Your entities are missing required attributes.
   * Your entities are using undefined attributes.
* Working with the glossary.
  * Uploading terms.
  * Downloading individual or all terms.
* Working with classifications.
  * Classify one entity with multiple classifications.
  * Classify multiple entities with a single classification.
  * Remove classification ("declassify") from an entity.
* Working with relationships.
  * Able to create arbitrary relationships between entities.
  * e.g. associating a given column with a table.
  * Able to upload relationship definitions.
* Deleting types (by name) or entities (by guid).
* Search (only for Azure Purview advanced search).
* Authentication to Azure Purview via Service Principal.
* Authentication using basic authentication of username and password for open source Atlas.
"""

def setup_package():
    setuptools.setup(
        name="pyapacheatlas",
        version=__version__,
        author="Will Johnson",
        author_email="will@willj.com",
        description="A package to simplify working with the Apache Atlas REST APIs for Atlas and Azure Purview.",
        long_description_content_type="text/markdown",
        url="https://github.com/wjohnson/pyapacheatlas",
        packages=setuptools.find_packages(),
        install_requires=[
            'openpyxl>=3.0',
            'requests>=2.0'
        ],
        classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
        ],
        python_requires='>=3.6',
        long_description=LONG_DESCRIPTION
    )

if __name__ == "__main__":
    setup_package()


from functools import wraps
import warnings


class AtlasException(BaseException):
    pass


def PurviewOnly(func):
    """
    Raise a runtime warning if you are using an AtlasClient (or non Purview)
    client. Intended to wrap specific client methods that are only available
    in Purview.
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        if not args[0].is_purview:
            warnings.warn(
                "You're using a Purview only feature on a non-purview endpoint.",
                RuntimeWarning
            )
        return func(*args, **kwargs)
    return wrapper


def PurviewLimitation(func):
    """
    Raise a runtime warning if you are using a PurviewClient. Intended to wrap
    specific client methods that have limitations due to Purview.
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        if args[0].is_purview:
            warnings.warn(
                "You're using a feature that Purview does not implement fully.",
                RuntimeWarning
            )
        return func(*args, **kwargs)
    return wrapper


class GuidTracker():
    """
    Always grab the next available guid by either inrementing or decrementing.
    When defining an interconnected set of Atlas Entities, you use a
    negative integer to provide an entity with a temporary unique id.
    """

    def __init__(self, starting=-1000, direction="decrease"):
        """
        :param int starting: A negative integer to start your guid tracker on.
        :param str direction: Either increase or decrease. It controls
            whether you increment or decrement the guid.
        """
        _ALLOWED_DIRECTIONS = ["increase", "decrease"]

        if direction not in _ALLOWED_DIRECTIONS:
            raise NotImplementedError(
                "The direction of {} is not supported.  Only {}".format(
                    direction, _ALLOWED_DIRECTIONS))

        self._guid = starting
        self._direction = direction

    def _decide_next_guid(self):
        """
        Do the math to determine what the next guid would be but do not
        update the guid.

        :return: The next guid you would receive.
        :rtype: int
        """
        return self._guid + (-1 if self._direction == "decrease" else 1)

    def get_guid(self):
        """
        Retrieve the next unique guid and update the guid.

        :return: A "unique" integer guid for your atlas object.
        :rtype: int
        """
        self._guid = self._decide_next_guid()
        return self._guid

    def peek_next_guid(self):
        """
        Peek at the next guid without updating the guid.

        :return: The next guid you would receive.
        :rtype: int
        """
        return self._decide_next_guid()
